# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import asyncio
import json
import os
import shutil
import sys
from pathlib import Path
from typing import Dict

import pytest
import yaml

from otx.cli.tools.find import SUPPORTED_BACKBONE_BACKENDS as find_supported_backends
from otx.cli.tools.find import SUPPORTED_TASKS as find_supported_tasks
from otx.cli.utils.nncf import get_number_of_fakequantizers_in_xml


def get_template_rel_dir(template):
    return os.path.dirname(os.path.relpath(template.model_template_path))


def get_template_dir(template, root) -> str:

    # Get the template directory of the algorithm.
    # The location of the template files are as follows:
    # ~/training_extensions/otx/algorithms/<algorithm>/**/template.yaml
    # To get the ``algorithm``, index of the "algorithms" can be
    # searched, where ``algorithm`` comes next.
    template_path_parts = template.model_template_path.split(os.sep)
    idx = template_path_parts.index("algorithms")
    algorithm = template_path_parts[idx + 1]

    algo_backend_dir = f"otx/algorithms/{algorithm}"
    work_dir = os.path.join(root, f"otx/algorithms/{algorithm}")
    template_dir = os.path.dirname(os.path.relpath(template.model_template_path, start=algo_backend_dir))
    template_work_dir = os.path.join(work_dir, template_dir)

    os.makedirs(template_work_dir, exist_ok=True)
    return template_work_dir


def runner(
    cmd,
    stdout_stream=sys.stdout.buffer,
    stderr_stream=sys.stderr.buffer,
    **kwargs,
):
    async def stream_handler(in_stream, out_stream):
        output = bytearray()
        # buffer line
        line = bytearray()
        while True:
            c = await in_stream.read(1)
            if not c:
                break
            line.extend(c)
            if c == b"\n":
                out_stream.write(line)
                output.extend(line)
                line = bytearray()
        return output

    async def run_and_capture(cmd):
        environ = os.environ.copy()
        environ["PYTHONUNBUFFERED"] = "1"
        process = await asyncio.create_subprocess_exec(
            *cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            env=environ,
            **kwargs,
        )

        try:
            stdout, stderr = await asyncio.gather(
                stream_handler(process.stdout, stdout_stream),
                stream_handler(process.stderr, stderr_stream),
            )
        except Exception:
            process.kill()
            raise
        finally:
            rc = await process.wait()
        return rc, stdout, stderr

    rc, stdout, stderr = asyncio.run(run_and_capture(cmd))

    return rc, stdout, stderr


def check_run(cmd, **kwargs):
    rc, _, stderr = runner(cmd, **kwargs)
    sys.stdout.flush()
    sys.stderr.flush()

    if rc != 0:
        stderr = stderr.decode("utf-8").splitlines()
        i = 0
        for i, line in enumerate(stderr):
            if line.startswith("Traceback"):
                break
        stderr = "\n".join(stderr[i:])
    assert rc == 0, stderr


def otx_train_testing(template, root, otx_dir, args):
    template_work_dir = get_template_dir(template, root)
    command_line = ["otx", "train", template.model_template_path]
    for arg in [
        "--train-ann_file",
        "--train-data-roots",
        "--val-ann-file",
        "--val-data-roots",
        "--unlabeled-data-roots",
        "--unlabeled-file-list",
    ]:
        arg_value = args.get(arg, None)
        if arg_value:
            command_line.extend([arg, os.path.join(otx_dir, arg_value)])
    command_line.extend(["--save-model-to", f"{template_work_dir}/trained_{template.model_template_id}"])
    command_line.extend(["--work-dir", f"{template_work_dir}"])
    if "--gpus" in args:
        command_line.extend(["--gpus", args["--gpus"]])
        if "--multi-gpu-port" in args:
            command_line.extend(["--multi-gpu-port", args["--multi-gpu-port"]])
    command_line.extend(args["train_params"])
    check_run(command_line)
    assert os.path.exists(f"{template_work_dir}/trained_{template.model_template_id}/weights.pth")
    assert os.path.exists(f"{template_work_dir}/trained_{template.model_template_id}/label_schema.json")


def otx_resume_testing(template, root, otx_dir, args):
    template_work_dir = get_template_dir(template, root)
    command_line = [
        "otx",
        "train",
        template.model_template_path,
    ]
    for option in [
        "--train-ann-file",
        "--train-data-roots",
        "--val-ann-file",
        "--val-data-roots",
        "--unlabeled-data-roots",
        "--unlabeled-file-list",
        "--resume-from",
    ]:
        if option in args:
            command_line.extend([option, f"{os.path.join(otx_dir, args[option])}"])

    command_line.extend(["--save-model-to", f"{template_work_dir}/trained_for_resume_{template.model_template_id}"])
    command_line.extend(["--work-dir", f"{template_work_dir}"])
    command_line.extend(args["train_params"])
    check_run(command_line)
    assert os.path.exists(f"{template_work_dir}/trained_for_resume_{template.model_template_id}/weights.pth")
    assert os.path.exists(f"{template_work_dir}/trained_for_resume_{template.model_template_id}/label_schema.json")


def otx_hpo_testing(template, root, otx_dir, args):
    template_work_dir = get_template_dir(template, root)
    if os.path.exists(f"{template_work_dir}/hpo"):
        shutil.rmtree(f"{template_work_dir}/hpo")

    command_line = ["otx", "train", template.model_template_path]

    for arg in ["--train-data-roots", "--val-data-roots"]:
        arg_value = args.get(arg, None)
        if arg_value:
            command_line.extend([arg, os.path.join(otx_dir, arg_value)])
    command_line.extend(["--save-model-to", f"{template_work_dir}/hpo_trained_{template.model_template_id}"])
    command_line.extend(["--work-dir", f"{template_work_dir}"])
    command_line.extend(["--enable-hpo", "--hpo-time-ratio", "1"])

    command_line.extend(args["train_params"])
    check_run(command_line)
    trials_json = list(
        filter(lambda x: x.name.split(".")[0].isnumeric(), Path(f"{template_work_dir}/hpo/").rglob("*.json"))
    )
    assert trials_json
    for trial_json in trials_json:
        with trial_json.open("r") as f:
            trial_result = json.load(f)
        assert trial_result.get("score")

    assert os.path.exists(f"{template_work_dir}/hpo_trained_{template.model_template_id}/weights.pth")
    assert os.path.exists(f"{template_work_dir}/hpo_trained_{template.model_template_id}/label_schema.json")


def otx_export_testing(template, root):
    template_work_dir = get_template_dir(template, root)
    command_line = [
        "otx",
        "export",
        template.model_template_path,
        "--load-weights",
        f"{template_work_dir}/trained_{template.model_template_id}/weights.pth",
        "--save-model-to",
        f"{template_work_dir}/exported_{template.model_template_id}",
        "--dump_features",
    ]
    check_run(command_line)
    assert os.path.exists(f"{template_work_dir}/exported_{template.model_template_id}/openvino.xml")
    assert os.path.exists(f"{template_work_dir}/exported_{template.model_template_id}/openvino.bin")
    assert os.path.exists(f"{template_work_dir}/exported_{template.model_template_id}/label_schema.json")


def otx_eval_testing(template, root, otx_dir, args):
    template_work_dir = get_template_dir(template, root)

    command_line = [
        "otx",
        "eval",
        template.model_template_path,
        "--test-data-roots",
        f'{os.path.join(otx_dir, args["--test-data-roots"])}',
        "--load-weights",
        f"{template_work_dir}/trained_{template.model_template_id}/weights.pth",
        "--save-performance",
        f"{template_work_dir}/trained_{template.model_template_id}/performance.json",
    ]
    command_line.extend(["--work-dir", f"{template_work_dir}"])
    command_line.extend(args.get("eval_params", []))
    check_run(command_line)
    assert os.path.exists(f"{template_work_dir}/trained_{template.model_template_id}/performance.json")


def otx_eval_openvino_testing(template, root, otx_dir, args, threshold):
    template_work_dir = get_template_dir(template, root)
    command_line = [
        "otx",
        "eval",
        template.model_template_path,
        "--test-data-roots",
        f'{os.path.join(otx_dir, args["--test-data-roots"])}',
        "--load-weights",
        f"{template_work_dir}/exported_{template.model_template_id}/openvino.xml",
        "--save-performance",
        f"{template_work_dir}/exported_{template.model_template_id}/performance.json",
    ]
    command_line.extend(["--work-dir", f"{template_work_dir}"])
    check_run(command_line)
    assert os.path.exists(f"{template_work_dir}/exported_{template.model_template_id}/performance.json")
    with open(f"{template_work_dir}/trained_{template.model_template_id}/performance.json") as read_file:
        trained_performance = json.load(read_file)
    with open(f"{template_work_dir}/exported_{template.model_template_id}/performance.json") as read_file:
        exported_performance = json.load(read_file)

    for k in trained_performance.keys():
        assert (
            exported_performance[k] >= trained_performance[k]
            or abs(trained_performance[k] - exported_performance[k]) / (trained_performance[k] + 1e-10) <= threshold
        ), f"{trained_performance[k]=}, {exported_performance[k]=}"


def otx_demo_testing(template, root, otx_dir, args):
    template_work_dir = get_template_dir(template, root)
    command_line = [
        "otx",
        "demo",
        template.model_template_path,
        "--load-weights",
        f"{template_work_dir}/trained_{template.model_template_id}/weights.pth",
        "--input",
        os.path.join(otx_dir, args["--input"]),
        "--delay",
        "-1",
    ]
    check_run(command_line)


def otx_demo_openvino_testing(template, root, otx_dir, args):
    template_work_dir = get_template_dir(template, root)
    command_line = [
        "otx",
        "demo",
        template.model_template_path,
        "--load-weights",
        f"{template_work_dir}/exported_{template.model_template_id}/openvino.xml",
        "--input",
        os.path.join(otx_dir, args["--input"]),
        "--delay",
        "-1",
    ]
    check_run(command_line)


def otx_deploy_openvino_testing(template, root, otx_dir, args):
    template_work_dir = get_template_dir(template, root)
    deployment_dir = f"{template_work_dir}/deployed_{template.model_template_id}"
    command_line = [
        "otx",
        "deploy",
        template.model_template_path,
        "--load-weights",
        f"{template_work_dir}/exported_{template.model_template_id}/openvino.xml",
        "--save-model-to",
        deployment_dir,
    ]
    check_run(command_line)
    check_run(["unzip", "-o", "openvino.zip"], cwd=deployment_dir)
    # TODO: Need to check Requirements.txt & new environment is working
    # check_run(
    #     ["python3", "-m", "venv", "venv"],
    #     cwd=os.path.join(deployment_dir, "python"),
    # )
    # check_run(
    #     ["python3", "-m", "pip", "install", "wheel"],
    #     cwd=os.path.join(deployment_dir, "python"),
    # )
    # check_run(
    #     ["python3", "-m", "pip", "install", "pip", "--upgrade"],
    #     cwd=os.path.join(deployment_dir, "python"),
    # )
    # check_run(
    #     ["python3", "-m", "pip", "install", "torch>=1.8.1, <=1.9.1"],
    #     cwd=os.path.join(deployment_dir, "python"),
    # )
    # check_run(
    #     [
    #         "python3",
    #         "-m",
    #         "pip",
    #         "install",
    #         "-r",
    #         os.path.join(deployment_dir, "python", "requirements.txt"),
    #     ],
    #     cwd=os.path.join(deployment_dir, "python"),
    # )
    check_run(
        [
            "python3",
            "demo.py",
            "-m",
            "../model",
            "-i",
            os.path.join(otx_dir, args["--input"]),
            "--no_show",
        ],
        cwd=os.path.join(deployment_dir, "python"),
    )


def otx_eval_deployment_testing(template, root, otx_dir, args, threshold):
    template_work_dir = get_template_dir(template, root)
    command_line = [
        "otx",
        "eval",
        template.model_template_path,
        "--test-data-roots",
        f'{os.path.join(otx_dir, args["--test-data-roots"])}',
        "--load-weights",
        f"{template_work_dir}/deployed_{template.model_template_id}/openvino.zip",
        "--save-performance",
        f"{template_work_dir}/deployed_{template.model_template_id}/performance.json",
    ]
    command_line.extend(["--work-dir", f"{template_work_dir}"])
    check_run(command_line)
    assert os.path.exists(f"{template_work_dir}/deployed_{template.model_template_id}/performance.json")
    with open(f"{template_work_dir}/exported_{template.model_template_id}/performance.json") as read_file:
        exported_performance = json.load(read_file)
    with open(f"{template_work_dir}/deployed_{template.model_template_id}/performance.json") as read_file:
        deployed_performance = json.load(read_file)

    for k in exported_performance.keys():
        assert (
            deployed_performance[k] >= exported_performance[k]
            or abs(exported_performance[k] - deployed_performance[k]) / (exported_performance[k] + 1e-10) <= threshold
        ), f"{exported_performance[k]=}, {deployed_performance[k]=}"


def otx_demo_deployment_testing(template, root, otx_dir, args):
    template_work_dir = get_template_dir(template, root)
    command_line = [
        "otx",
        "demo",
        template.model_template_path,
        "--load-weights",
        f"{template_work_dir}/deployed_{template.model_template_id}/openvino.zip",
        "--input",
        os.path.join(otx_dir, args["--input"]),
        "--delay",
        "-1",
    ]
    check_run(command_line)


def pot_optimize_testing(template, root, otx_dir, args):
    template_work_dir = get_template_dir(template, root)
    command_line = [
        "otx",
        "optimize",
        template.model_template_path,
        "--train-data-roots",
        f'{os.path.join(otx_dir, args["--train-data-roots"])}',
        "--val-data-roots",
        f'{os.path.join(otx_dir, args["--val-data-roots"])}',
        "--load-weights",
        f"{template_work_dir}/exported_{template.model_template_id}/openvino.xml",
        "--save-model-to",
        f"{template_work_dir}/pot_{template.model_template_id}",
    ]
    command_line.extend(["--work-dir", f"{template_work_dir}"])
    check_run(command_line)
    assert os.path.exists(f"{template_work_dir}/pot_{template.model_template_id}/openvino.xml")
    assert os.path.exists(f"{template_work_dir}/pot_{template.model_template_id}/openvino.bin")
    assert os.path.exists(f"{template_work_dir}/pot_{template.model_template_id}/label_schema.json")


def _validate_fq_in_xml(xml_path, path_to_ref_data, compression_type, test_name):
    num_fq = get_number_of_fakequantizers_in_xml(xml_path)
    assert os.path.exists(path_to_ref_data), f"Reference file does not exist: {path_to_ref_data} [num_fq = {num_fq}]"

    with open(path_to_ref_data, encoding="utf-8") as stream:
        ref_data = yaml.safe_load(stream)
    ref_num_fq = ref_data.get(test_name, {}).get(compression_type, {}).get("number_of_fakequantizers", -1)
    assert num_fq == ref_num_fq, f"Incorrect number of FQs in optimized model: {num_fq} != {ref_num_fq}"


def pot_validate_fq_testing(template, root, otx_dir, task_type, test_name):
    template_work_dir = get_template_dir(template, root)
    xml_path = f"{template_work_dir}/pot_{template.model_template_id}/openvino.xml"
    path_to_ref_data = os.path.join(
        otx_dir, "tests", "e2e/cli", task_type, "reference", template.model_template_id, "compressed_model.yml"
    )
    _validate_fq_in_xml(xml_path, path_to_ref_data, "pot", test_name)


def pot_eval_testing(template, root, otx_dir, args):
    template_work_dir = get_template_dir(template, root)
    command_line = [
        "otx",
        "eval",
        template.model_template_path,
        "--test-data-roots",
        f'{os.path.join(otx_dir, args["--test-data-roots"])}',
        "--load-weights",
        f"{template_work_dir}/pot_{template.model_template_id}/openvino.xml",
        "--save-performance",
        f"{template_work_dir}/pot_{template.model_template_id}/performance.json",
    ]
    command_line.extend(["--work-dir", f"{template_work_dir}"])
    check_run(command_line)
    assert os.path.exists(f"{template_work_dir}/pot_{template.model_template_id}/performance.json")


def nncf_optimize_testing(template, root, otx_dir, args):
    template_work_dir = get_template_dir(template, root)
    command_line = [
        "otx",
        "optimize",
        template.model_template_path,
        "--train-data-roots",
        f'{os.path.join(otx_dir, args["--train-data-roots"])}',
        "--val-data-roots",
        f'{os.path.join(otx_dir, args["--val-data-roots"])}',
        "--load-weights",
        f"{template_work_dir}/trained_{template.model_template_id}/weights.pth",
        "--save-model-to",
        f"{template_work_dir}/nncf_{template.model_template_id}",
        "--save-performance",
        f"{template_work_dir}/nncf_{template.model_template_id}/train_performance.json",
    ]
    command_line.extend(["--work-dir", f"{template_work_dir}"])
    command_line.extend(args["train_params"])
    check_run(command_line)
    assert os.path.exists(f"{template_work_dir}/nncf_{template.model_template_id}/weights.pth")
    assert os.path.exists(f"{template_work_dir}/nncf_{template.model_template_id}/label_schema.json")


def nncf_export_testing(template, root):
    template_work_dir = get_template_dir(template, root)
    command_line = [
        "otx",
        "export",
        template.model_template_path,
        "--load-weights",
        f"{template_work_dir}/nncf_{template.model_template_id}/weights.pth",
        "--save-model-to",
        f"{template_work_dir}/exported_nncf_{template.model_template_id}",
    ]
    check_run(command_line)
    assert os.path.exists(f"{template_work_dir}/exported_nncf_{template.model_template_id}/openvino.xml")
    assert os.path.exists(f"{template_work_dir}/exported_nncf_{template.model_template_id}/openvino.bin")
    assert os.path.exists(f"{template_work_dir}/exported_nncf_{template.model_template_id}/label_schema.json")
    original_bin_size = os.path.getsize(f"{template_work_dir}/exported_{template.model_template_id}/openvino.bin")
    compressed_bin_size = os.path.getsize(
        f"{template_work_dir}/exported_nncf_{template.model_template_id}/openvino.bin"
    )
    assert compressed_bin_size < original_bin_size, f"{compressed_bin_size=}, {original_bin_size=}"


def nncf_validate_fq_testing(template, root, otx_dir, task_type, test_name):
    template_work_dir = get_template_dir(template, root)
    xml_path = f"{template_work_dir}/exported_nncf_{template.model_template_id}/openvino.xml"
    path_to_ref_data = os.path.join(
        otx_dir, "tests", "e2e/cli", task_type, "reference", template.model_template_id, "compressed_model.yml"
    )

    _validate_fq_in_xml(xml_path, path_to_ref_data, "nncf", test_name)


def nncf_eval_testing(template, root, otx_dir, args, threshold):
    template_work_dir = get_template_dir(template, root)
    command_line = [
        "otx",
        "eval",
        template.model_template_path,
        "--test-data-roots",
        f'{os.path.join(otx_dir, args["--test-data-roots"])}',
        "--load-weights",
        f"{template_work_dir}/nncf_{template.model_template_id}/weights.pth",
        "--save-performance",
        f"{template_work_dir}/nncf_{template.model_template_id}/performance.json",
    ]
    command_line.extend(["--work-dir", f"{template_work_dir}"])
    check_run(command_line)
    assert os.path.exists(f"{template_work_dir}/nncf_{template.model_template_id}/performance.json")
    with open(f"{template_work_dir}/nncf_{template.model_template_id}/train_performance.json") as read_file:
        trained_performance = json.load(read_file)
    with open(f"{template_work_dir}/nncf_{template.model_template_id}/performance.json") as read_file:
        evaluated_performance = json.load(read_file)

    for k in trained_performance.keys():
        assert (
            evaluated_performance[k] >= trained_performance[k]
            or abs(trained_performance[k] - evaluated_performance[k]) / (trained_performance[k] + 1e-10) <= threshold
        ), f"{trained_performance[k]=}, {evaluated_performance[k]=}"


def nncf_eval_openvino_testing(template, root, otx_dir, args):
    template_work_dir = get_template_dir(template, root)
    command_line = [
        "otx",
        "eval",
        template.model_template_path,
        "--test-data-roots",
        f'{os.path.join(otx_dir, args["--test-data-roots"])}',
        "--load-weights",
        f"{template_work_dir}/exported_nncf_{template.model_template_id}/openvino.xml",
        "--save-performance",
        f"{template_work_dir}/exported_nncf_{template.model_template_id}/performance.json",
    ]
    command_line.extend(["--work-dir", f"{template_work_dir}"])
    check_run(command_line)
    assert os.path.exists(f"{template_work_dir}/exported_nncf_{template.model_template_id}/performance.json")


def xfail_templates(templates, xfail_template_ids_reasons):
    xfailed_templates = []
    for template in templates:
        reasons = [
            reason for template_id, reason in xfail_template_ids_reasons if template_id == template.model_template_id
        ]
        if len(reasons) == 0:
            xfailed_templates.append(template)
        elif len(reasons) == 1:
            xfailed_templates.append(pytest.param(template, marks=pytest.mark.xfail(reason=reasons[0])))
        else:
            raise RuntimeError(
                "More than one reason for template. If you have more than one Jira tickets, list them in one reason."
            )
    return xfailed_templates


def otx_explain_testing(template, root, otx_dir, args):
    template_work_dir = get_template_dir(template, root)
    if "RCNN" in template.model_template_id:
        test_algorithm = "ActivationMap"
    else:
        test_algorithm = "ClassWiseSaliencyMap"

    train_ann_file = args.get("--train-ann-file", "")
    if "hierarchical" in train_ann_file:
        train_type = "hierarchical"
    elif "multilabel" in train_ann_file:
        train_type = "multilabel"
    else:
        train_type = "default"

    save_dir = f"explain_{template.model_template_id}/{test_algorithm}/{train_type}/"
    output_dir = os.path.join(template_work_dir, save_dir)
    command_line = [
        "otx",
        "explain",
        template.model_template_path,
        "--load-weights",
        f"{template_work_dir}/trained_{template.model_template_id}/weights.pth",
        "--explain-data-root",
        os.path.join(otx_dir, args["--input"]),
        "--save-explanation-to",
        output_dir,
        "--explain-algorithm",
        test_algorithm,
    ]
    check_run(command_line)
    assert os.path.exists(output_dir)
    assert len(os.listdir(output_dir)) > 0


def otx_explain_openvino_testing(template, root, otx_dir, args):
    template_work_dir = get_template_dir(template, root)
    if "RCNN" in template.model_template_id:
        test_algorithm = "ActivationMap"
    else:
        test_algorithm = "ClassWiseSaliencyMap"

    train_ann_file = args.get("--train-ann-file", "")
    if "hierarchical" in train_ann_file:
        train_type = "hierarchical"
    elif "multilabel" in train_ann_file:
        train_type = "multilabel"
    else:
        train_type = "default"

    save_dir = f"explain_{template.model_template_id}/{test_algorithm}/{train_type}/"
    output_dir = os.path.join(template_work_dir, save_dir)
    command_line = [
        "otx",
        "explain",
        template.model_template_path,
        "--load-weights",
        f"{template_work_dir}/exported_{template.model_template_id}/openvino.xml",
        "--explain-data-root",
        os.path.join(otx_dir, args["--input"]),
        "--save-explanation-to",
        output_dir,
        "--explain-algorithm",
        test_algorithm,
    ]
    check_run(command_line)
    assert os.path.exists(output_dir)
    assert len(os.listdir(output_dir)) > 0


def otx_find_testing():
    """Performs several options of available otx find."""
    # Find all model template
    command_line = ["otx", "find", "--template"]
    check_run(command_line)

    # Find command per tasks
    for task in find_supported_tasks:
        command_line = ["otx", "find", "--template", "--task", task]
        check_run(command_line)

    # Find Backbones per backends
    for backbone_backends in find_supported_backends:
        command_line = [
            "otx",
            "find",
            "--backbone",
            backbone_backends,
        ]
        check_run(command_line)


def otx_build_task_testing(root, task):
    """Build OTX-workspace per tasks.

    Build and verify the otx-workspace corresponding to each task.
    """
    # Build otx-workspace per tasks check - Default Model Template only
    command_line = [
        "otx",
        "build",
        "--task",
        task,
        "--work-dir",
        os.path.join(root, f"otx-workspace-{task}"),
    ]
    check_run(command_line)


def otx_build_backbone_testing(root, backbone_args):
    """Build backbone & Update model testing.

    Build each backbone to create backbone.yaml into workspace,
    build for the default model for the task,
    and even test updating the model config.
    This is done on the premise that the otx_workspace
    has been created well through otx_build_task_testing.
    """
    task, backbone = backbone_args
    task_workspace = os.path.join(root, f"otx-workspace-{task}")
    command_line = [
        "otx",
        "build",
        "--task",
        f"{task}",
        "--work-dir",
        task_workspace,
    ]
    check_run(command_line)
    assert os.path.exists(task_workspace)

    # Build model.py from backbone type
    command_line = [
        "otx",
        "build",
        "--backbone",
        backbone,
        "--work-dir",
        task_workspace,
    ]
    check_run(command_line)
    from otx.mpa.utils.config_utils import MPAConfig

    model_config = MPAConfig.fromfile(os.path.join(task_workspace, "model.py"))
    assert os.path.exists(os.path.join(task_workspace, "model.py"))
    assert "backbone" in model_config["model"], "'backbone' is not in model configs"
    assert (
        model_config["model"]["backbone"]["type"] == backbone
    ), f"{model_config['model']['backbone']['type']} != {backbone}"


def otx_build_testing(root, args: Dict[str, str], expected: Dict[str, str]):
    workspace_root = os.path.join(root, "otx-workspace")
    command_line = ["otx", "build", "--work-dir", workspace_root]
    for option, value in args.items():
        command_line.extend([option, value])
    check_run(command_line)
    from otx.mpa.utils.config_utils import MPAConfig

    template_config = MPAConfig.fromfile(os.path.join(workspace_root, "template.yaml"))
    assert template_config.name == expected["model"]
    assert (
        template_config.hyper_parameters.parameter_overrides.algo_backend.train_type.default_value
        == expected["train_type"].upper()
    )


def otx_build_auto_config(root, otx_dir: str, args: Dict[str, str]):
    workspace_root = os.path.join(root, "otx-workspace")
    command_line = ["otx", "build", "--work-dir", workspace_root]

    for option, value in args.items():
        if option in ["--train-data-roots", "--val-data-roots"]:
            command_line.extend([option, f"{os.path.join(otx_dir, value)}"])
        elif option in ["--task"]:
            command_line.extend([option, args[option]])
    check_run(command_line)


def otx_train_auto_config(root, otx_dir: str, args: Dict[str, str]):
    work_dir = os.path.join(root, "otx-workspace")
    command_line = ["otx", "train"]

    for option, value in args.items():
        if option == "template":
            command_line.extend([args[option]])
        elif option in ["--train-data-roots", "--val-data-roots"]:
            command_line.extend([option, f"{os.path.join(otx_dir, value)}"])
    command_line.extend(["--save-model-to", f"{work_dir}"])
    command_line.extend(["--work-dir", f"{work_dir}"])
    command_line.extend(args["train_params"])
    check_run(command_line)



"""Model templates registry."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
import glob
import os
from pathlib import Path
from typing import Optional

import yaml

from otx.api.entities.model_template import parse_model_template
from otx.cli.utils.importing import get_backbone_list, get_otx_root_path


class Registry:
    """Class that implements a model templates registry."""

    def __init__(self, templates_dir=None, templates=None, experimental=False):
        if templates is None:
            if templates_dir is None:
                templates_dir = os.getenv("TEMPLATES_DIR")

            if templates_dir is None:
                raise RuntimeError("The templates_dir is not set.")

            template_filenames = glob.glob(os.path.join(templates_dir, "**", "template.yaml"), recursive=True)
            if experimental:
                template_filenames.extend(
                    glob.glob(
                        os.path.join(templates_dir, "**", "template_experimental.yaml"),
                        recursive=True,
                    )
                )
            template_filenames = [os.path.abspath(p) for p in template_filenames]

            self.templates = []

            for template_file in template_filenames:
                self.templates.append(parse_model_template(template_file))
        else:
            self.templates = copy.deepcopy(templates)

        self.task_types = self.__collect_task_types(self.templates)

    @staticmethod
    def __collect_task_types(templates):
        return {template.task_type for template in templates}

    def filter(self, framework=None, task_type=None):
        """Filters registry by framework and/or task type and returns filtered registry."""

        templates = copy.deepcopy(self.templates)
        if framework is not None:
            templates = [template for template in templates if template.framework.lower() == framework.lower()]
        if task_type is not None:
            templates = [template for template in templates if str(template.task_type).lower() == task_type.lower()]
        return Registry(templates=templates)

    def get(self, template_id, skip_error=False):
        """Returns a model template with specified template_id or template.name."""

        templates = [
            template for template in self.templates if template_id in (template.model_template_id, template.name)
        ]
        if not templates:
            if skip_error:
                return None
            raise ValueError(f"Could not find a template with {template_id} in registry.")
        return templates[0]

    def get_backbones(self, backend_list):
        """Returns list of backbones for a given template."""
        backbone_list = {}
        for backend in backend_list:
            backbone_list[backend] = get_backbone_list(backend)
        return backbone_list

    def __str__(self):
        """Returns the string representation of the registry."""
        templates_infos = [
            {
                "name": t.name,
                "id": t.model_template_id,
                "path": t.model_template_path,
                "task_type": str(t.task_type),
            }
            for t in self.templates
        ]
        return yaml.dump(templates_infos)


def find_and_parse_model_template(path_or_id):
    """In first function attempts to read a model template from disk under assumption that a path is passed.

    If the attempt is failed, it tries to find template in registry under assumption that an ID or name is passed.
    """
    # Return None Type
    if not path_or_id:
        return path_or_id

    # 1. Find from path
    if is_template(path_or_id):
        return parse_model_template(path_or_id)
    # 2. Find from id or Name
    return Registry(get_otx_root_path()).get(path_or_id, skip_error=True)


def is_template(template_path: Optional[str]) -> bool:
    """A function that determines whether the corresponding template path is a template.

    Args:
        template_path (str): The path of the file you want to know if it is a template.

    Returns:
        bool: True if template_path is template file else False.
    """
    if template_path and Path(template_path).is_file() and "template" in Path(template_path).name:
        return True
    return False



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import tempfile

import pytest
from mmcv.utils import Config

from otx.algorithms.classification.adapters.mmcls.utils import (
    patch_config,
    patch_evaluation,
)
from otx.algorithms.common.adapters.mmcv.utils import get_dataset_configs
from otx.api.entities.id import ID
from otx.api.entities.label import Domain, LabelEntity
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@pytest.fixture
def otx_default_cls_config():
    config_name = "otx/algorithms/classification/configs/efficientnet_b0_cls_incr/model.py"
    conf = Config.fromfile(config_name)
    conf.checkpoint_config = {}
    conf.evaluation = {"metric": "None"}
    conf.data = {}
    return conf


@pytest.fixture
def otx_default_labels():
    return [
        LabelEntity(name=name, domain=Domain.CLASSIFICATION, is_empty=False, id=ID(i))
        for i, name in enumerate(["a", "b"])
    ]


@e2e_pytest_unit
def test_patch_config(otx_default_cls_config, otx_default_labels) -> None:
    """Test patch_config function.

    <Steps>
        1. Check work_dir
        2. Check removed high level pipelines
        3. Check checkpoint config update
        4. Check dataset labels
    """

    with tempfile.TemporaryDirectory() as work_dir:
        patch_config(otx_default_cls_config, work_dir, otx_default_labels)
        assert otx_default_cls_config.work_dir == work_dir

        assert otx_default_cls_config.get("train_pipeline", None) is None
        assert otx_default_cls_config.get("test_pipeline", None) is None
        assert otx_default_cls_config.get("train_pipeline_strong", None) is None

        assert otx_default_cls_config.checkpoint_config.max_keep_ckpts > 0
        assert otx_default_cls_config.checkpoint_config.interval > 0

        for subset in ("train", "val", "test"):
            for cfg in get_dataset_configs(otx_default_cls_config, subset):
                assert len(cfg.labels) == len(otx_default_labels)


@e2e_pytest_unit
def test_patch_evaluation(otx_default_cls_config) -> None:
    """Test patch_evaluation function.

    <Steps>
        1. Check eval metrics not empty
        2. Check the early stop metric is in eval metrics
        3. Check an exception is thrown for a frong task type
    """
    tasks = ["multilabel", "hierarchical", "normal"]
    for task in tasks:
        patch_evaluation(otx_default_cls_config, task)
        eval_metrics = otx_default_cls_config.evaluation.metric
        assert len(eval_metrics) > 0
        assert otx_default_cls_config.early_stop_metric in eval_metrics

    with pytest.raises(NotImplementedError):
        patch_evaluation(otx_default_cls_config, "None")



"""Utils for checking functions and methods arguments."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import inspect
import itertools
import typing
from abc import ABC, abstractmethod
from collections.abc import Sequence
from functools import wraps
from os.path import exists, splitext

import yaml
from numpy import floating
from omegaconf import DictConfig

IMAGE_FILE_EXTENSIONS = [
    ".bmp",
    ".dib",
    ".jpeg",
    ".jpg",
    ".jpe",
    ".jp2",
    ".png",
    ".webp",
    ".pbm",
    ".pgm",
    ".ppm",
    ".pxm",
    ".pnm",
    ".sr",
    ".ras",
    ".tiff",
    ".tif",
    ".exr",
    ".hdr",
    ".pic",
]


def get_bases(parameter) -> set:
    """Function to get set of all base classes of parameter."""

    def __get_bases(parameter_type):
        return [parameter_type.__name__] + list(
            itertools.chain.from_iterable(__get_bases(t1) for t1 in parameter_type.__bases__)
        )

    return set(__get_bases(type(parameter)))


def get_parameter_repr(parameter) -> str:
    """Function to get parameter representation."""
    try:
        parameter_str = repr(parameter)
    # pylint: disable=broad-except
    except Exception:
        parameter_str = "<unable to get parameter repr>"
    return parameter_str


def raise_value_error_if_parameter_has_unexpected_type(parameter, parameter_name, expected_type):
    """Function raises ValueError exception if parameter has unexpected type."""
    if isinstance(expected_type, typing.ForwardRef):
        expected_type = expected_type.__forward_arg__
    if isinstance(expected_type, str):
        parameter_types = get_bases(parameter)
        if not any(t == expected_type for t in parameter_types):
            parameter_str = get_parameter_repr(parameter)
            raise ValueError(
                f"Unexpected type of '{parameter_name}' parameter, expected: {expected_type}, "
                f"actual value: {parameter_str}"
            )
        return
    if expected_type == float:
        expected_type = (int, float, floating)
    if not isinstance(parameter, expected_type):
        parameter_type = type(parameter)
        parameter_str = get_parameter_repr(parameter)
        raise ValueError(
            f"Unexpected type of '{parameter_name}' parameter, expected: {expected_type}, actual: {parameter_type}, "
            f"actual value: {parameter_str}"
        )


def check_nested_elements_type(iterable, parameter_name, expected_type):
    """Function raises ValueError exception if one of elements in collection has unexpected type."""
    for element in iterable:
        check_parameter_type(
            parameter=element,
            parameter_name=f"nested {parameter_name}",
            expected_type=expected_type,
        )


def check_dictionary_keys_values_type(parameter, parameter_name, expected_key_class, expected_value_class):
    """Function raises ValueError exception if dictionary key or value has unexpected type."""
    for key, value in parameter.items():
        check_parameter_type(
            parameter=key,
            parameter_name=f"key in {parameter_name}",
            expected_type=expected_key_class,
        )
        check_parameter_type(
            parameter=value,
            parameter_name=f"value in {parameter_name}",
            expected_type=expected_value_class,
        )


def check_nested_classes_parameters(parameter, parameter_name, origin_class, nested_elements_class):
    """Function to check type of parameters with nested elements."""
    # Checking origin class
    raise_value_error_if_parameter_has_unexpected_type(
        parameter=parameter, parameter_name=parameter_name, expected_type=origin_class
    )
    # Checking nested elements
    if origin_class == dict:
        if len(nested_elements_class) != 2:
            raise TypeError("length of nested expected types for dictionary should be equal to 2")
        key, value = nested_elements_class
        check_dictionary_keys_values_type(
            parameter=parameter,
            parameter_name=parameter_name,
            expected_key_class=key,
            expected_value_class=value,
        )
    if origin_class in [list, set, tuple, Sequence]:
        if origin_class == tuple:
            tuple_length = len(nested_elements_class)
            if tuple_length > 2:
                raise NotImplementedError("length of nested expected types for Tuple should not exceed 2")
            if tuple_length == 2:
                if nested_elements_class[1] != Ellipsis:
                    raise NotImplementedError("expected homogeneous tuple annotation")
                nested_elements_class = nested_elements_class[0]
        else:
            if len(nested_elements_class) != 1:
                raise TypeError("length of nested expected types for Sequence should be equal to 1")
        check_nested_elements_type(
            iterable=parameter,
            parameter_name=parameter_name,
            expected_type=nested_elements_class,
        )


def check_parameter_type(parameter, parameter_name, expected_type):
    """Function extracts nested expected types and raises ValueError exception if parameter has unexpected type."""
    # pylint: disable=W0212
    if expected_type in [typing.Any, inspect._empty]:  # type: ignore
        return
    if not isinstance(expected_type, typing._GenericAlias):  # type: ignore
        raise_value_error_if_parameter_has_unexpected_type(
            parameter=parameter,
            parameter_name=parameter_name,
            expected_type=expected_type,
        )
        return
    expected_type_dict = expected_type.__dict__
    origin_class = expected_type_dict.get("__origin__")
    nested_elements_class = expected_type_dict.get("__args__")
    # Union type with nested elements check
    if origin_class == typing.Union:
        expected_args = expected_type_dict.get("__args__")
        checks_counter = 0
        errors_counter = 0
        for expected_arg in expected_args:
            try:
                checks_counter += 1
                check_parameter_type(parameter, parameter_name, expected_arg)
            except ValueError:
                errors_counter += 1
        if errors_counter == checks_counter:
            actual_type = type(parameter)
            raise ValueError(
                f"Unexpected type of '{parameter_name}' parameter, expected: {expected_args}, "
                f"actual type: {actual_type}, actual value: {parameter}"
            )
    # Checking parameters with nested elements
    elif issubclass(origin_class, typing.Iterable):
        check_nested_classes_parameters(
            parameter=parameter,
            parameter_name=parameter_name,
            origin_class=origin_class,
            nested_elements_class=nested_elements_class,
        )


def check_input_parameters_type(custom_checks: typing.Optional[dict] = None):
    """Decorator to check input parameters type.

    Args:
        custom_checks: dictionary where key - name of parameter and
            value - custom check class
    """
    if custom_checks is None:
        custom_checks = {}

    def _check_input_parameters_type(function):
        @wraps(function)
        def validate(*args, **kwargs):
            # Forming expected types dictionary
            signature = inspect.signature(function)
            expected_types_map = signature.parameters
            if len(expected_types_map) < len(args):
                raise TypeError("Too many positional arguments")
            # Forming input parameters dictionary
            input_parameters_values_map = dict(zip(signature.parameters.keys(), args))
            for key, value in kwargs.items():
                if key in input_parameters_values_map:
                    raise TypeError(f"Duplication of the parameter {key} -- both in args and kwargs")
                input_parameters_values_map[key] = value
            # Checking input parameters type
            for parameter_name in expected_types_map:
                parameter = input_parameters_values_map.get(parameter_name)
                if parameter_name not in input_parameters_values_map:
                    default_value = expected_types_map.get(parameter_name).default
                    # pylint: disable=protected-access
                    if default_value != inspect._empty:  # type: ignore
                        parameter = default_value
                if parameter_name in custom_checks:
                    custom_check = custom_checks[parameter_name]
                    if custom_check is None:
                        continue
                    custom_check(parameter, parameter_name).check()
                else:
                    check_parameter_type(
                        parameter=parameter,
                        parameter_name=parameter_name,
                        expected_type=expected_types_map.get(parameter_name).annotation,
                    )
            return function(**input_parameters_values_map)

        return validate

    return _check_input_parameters_type


def check_file_extension(file_path: str, file_path_name: str, expected_extensions: list):
    """Function raises ValueError exception if file has unexpected extension."""
    file_extension = splitext(file_path)[1].lower()
    if file_extension not in expected_extensions:
        raise ValueError(
            f"Unexpected extension of {file_path_name} file. expected: {expected_extensions} actual: {file_extension}"
        )


def check_that_null_character_absents_in_string(parameter: str, parameter_name: str):
    r"""Function raises ValueError exception if null character: '\0' is specified in path to file."""
    if "\0" in parameter:
        raise ValueError(f"null char \\0 is specified in {parameter_name}: {parameter}")


def check_that_file_exists(file_path: str, file_path_name: str):
    """Function raises ValueError exception if file not exists."""
    if not exists(file_path):
        raise ValueError(f"File {file_path} specified in '{file_path_name}' parameter not exists")


def check_that_parameter_is_not_empty(parameter, parameter_name):
    """Function raises ValueError if parameter is empty."""
    if not parameter:
        raise ValueError(f"parameter {parameter_name} is empty")


def check_that_all_characters_printable(parameter, parameter_name, allow_crlf=False):
    """Function raises ValueError if one of string-parameter characters is not printable."""
    if not allow_crlf:
        all_characters_printable = all(c.isprintable() for c in parameter)
    else:
        all_characters_printable = all((c.isprintable() or c == "\n" or c == "\r") for c in parameter)
    if not all_characters_printable:
        raise ValueError(rf"parameter {parameter_name} has not printable symbols: {parameter}")


def check_is_parameter_like_dataset(parameter, parameter_name):
    """Checks if the parameter is like a dataset.

    Function raises ValueError exception if parameter does not have __len__, __getitem__ and get_subset attributes of
    DataSet-type object.
    """
    for expected_attribute in ("__len__", "__getitem__", "get_subset"):
        if not hasattr(parameter, expected_attribute):
            parameter_type = type(parameter)
            raise ValueError(
                f"parameter '{parameter_name}' is not like DatasetEntity, actual type: {parameter_type} which does "
                f"not have expected '{expected_attribute}' dataset attribute"
            )


def check_file_path(parameter, parameter_name, expected_file_extensions):
    """Function to check file path string objects."""
    raise_value_error_if_parameter_has_unexpected_type(
        parameter=parameter,
        parameter_name=parameter_name,
        expected_type=str,
    )
    check_that_parameter_is_not_empty(parameter=parameter, parameter_name=parameter_name)
    check_file_extension(
        file_path=parameter,
        file_path_name=parameter_name,
        expected_extensions=expected_file_extensions,
    )
    check_that_null_character_absents_in_string(parameter=parameter, parameter_name=parameter_name)
    check_that_all_characters_printable(parameter=parameter, parameter_name=parameter_name)
    check_that_file_exists(file_path=parameter, file_path_name=parameter_name)


def check_directory_path(parameter, parameter_name):
    """Function to check directory path string objects."""
    raise_value_error_if_parameter_has_unexpected_type(
        parameter=parameter,
        parameter_name=parameter_name,
        expected_type=str,
    )
    check_that_parameter_is_not_empty(parameter=parameter, parameter_name=parameter_name)
    check_that_null_character_absents_in_string(parameter=parameter, parameter_name=parameter_name)
    check_that_all_characters_printable(parameter=parameter, parameter_name=parameter_name)


class BaseInputArgumentChecker(ABC):
    """Abstract class to check input arguments."""

    @abstractmethod
    def check(self):
        """Abstract method to check input arguments."""
        raise NotImplementedError("The check is not implemented")


class InputConfigCheck(BaseInputArgumentChecker):
    """Class to check input config_parameters."""

    def __init__(self, parameter, parameter_name):
        self.parameter = parameter
        self.parameter_name = parameter_name

    def check(self):
        """Method raises ValueError exception if "input_config" parameter is not equal to expected."""
        raise_value_error_if_parameter_has_unexpected_type(
            parameter=self.parameter,
            parameter_name=self.parameter_name,
            expected_type=(str, DictConfig, dict),
        )
        check_that_parameter_is_not_empty(parameter=self.parameter, parameter_name=self.parameter_name)
        if isinstance(self.parameter, str):
            check_that_null_character_absents_in_string(parameter=self.parameter, parameter_name=self.parameter_name)
            # yaml-format string is specified
            if isinstance(yaml.safe_load(self.parameter), dict):
                check_that_all_characters_printable(
                    parameter=self.parameter,
                    parameter_name=self.parameter_name,
                    allow_crlf=True,
                )
            # Path to file is specified
            else:
                check_file_extension(
                    file_path=self.parameter,
                    file_path_name=self.parameter_name,
                    expected_extensions=[".yaml"],
                )
                check_that_all_characters_printable(parameter=self.parameter, parameter_name=self.parameter_name)
                check_that_file_exists(file_path=self.parameter, file_path_name=self.parameter_name)


class FilePathCheck(BaseInputArgumentChecker):
    """Class to check file_path-like parameters."""

    def __init__(self, parameter, parameter_name, expected_file_extension):
        self.parameter = parameter
        self.parameter_name = parameter_name
        self.expected_file_extensions = expected_file_extension

    def check(self):
        """Method raises ValueError exception if file path parameter is not equal to expected."""
        check_file_path(self.parameter, self.parameter_name, self.expected_file_extensions)


class OptionalFilePathCheck(BaseInputArgumentChecker):
    """Class to check optional file_path-like parameters."""

    def __init__(self, parameter, parameter_name, expected_file_extension):
        self.parameter = parameter
        self.parameter_name = parameter_name
        self.expected_file_extensions = expected_file_extension

    def check(self):
        """Method raises ValueError exception if file path parameter is not equal to expected."""
        if self.parameter is not None:
            check_file_path(self.parameter, self.parameter_name, self.expected_file_extensions)


class DatasetParamTypeCheck(BaseInputArgumentChecker):
    """Class to check DatasetEntity-type parameters."""

    def __init__(self, parameter, parameter_name):
        self.parameter = parameter
        self.parameter_name = parameter_name

    def check(self):
        """Method raises ValueError exception if parameter is not equal to Dataset."""
        check_is_parameter_like_dataset(parameter=self.parameter, parameter_name=self.parameter_name)


class OptionalImageFilePathCheck(OptionalFilePathCheck):
    """Class to check optional image file path parameters."""

    def __init__(self, parameter, parameter_name):
        super().__init__(
            parameter=parameter,
            parameter_name=parameter_name,
            expected_file_extension=IMAGE_FILE_EXTENSIONS,
        )


class YamlFilePathCheck(FilePathCheck):
    """Class to check optional yaml file path parameters."""

    def __init__(self, parameter, parameter_name):
        super().__init__(
            parameter=parameter,
            parameter_name=parameter_name,
            expected_file_extension=[".yaml"],
        )


class JsonFilePathCheck(FilePathCheck):
    """Class to check optional yaml file path parameters."""

    def __init__(self, parameter, parameter_name):
        super().__init__(
            parameter=parameter,
            parameter_name=parameter_name,
            expected_file_extension=[".json"],
        )


class DirectoryPathCheck(BaseInputArgumentChecker):
    """Class to check directory path parameters."""

    def __init__(self, parameter, parameter_name):
        self.parameter = parameter
        self.parameter_name = parameter_name

    def check(self):
        """Method raises ValueError exception if directory path parameter is not equal to expected."""
        check_directory_path(parameter=self.parameter, parameter_name=self.parameter_name)


class OptionalDirectoryPathCheck(BaseInputArgumentChecker):
    """Class to check optional directory path parameters."""

    def __init__(self, parameter, parameter_name):
        self.parameter = parameter
        self.parameter_name = parameter_name

    def check(self):
        """Method raises ValueError exception if directory path parameter is not equal to expected."""
        if self.parameter is not None:
            check_directory_path(parameter=self.parameter, parameter_name=self.parameter_name)



"""Sync pipeline executor based on ModelAPI."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import List, Tuple, Union

import numpy as np

from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.usecases.exportable_code.demo.demo_package.model_container import (
    ModelContainer,
)
from otx.api.usecases.exportable_code.demo.demo_package.utils import (
    create_output_converter,
)
from otx.api.usecases.exportable_code.streamer import get_streamer
from otx.api.usecases.exportable_code.visualizers import Visualizer
from otx.api.utils.shape_factory import ShapeFactory


class ChainExecutor:
    """Sync executor for task-chain inference.

    Args:
        models: list of models for inference
        visualizer: visualizer of inference results
    """

    def __init__(
        self,
        models: List[ModelContainer],
        visualizer: Visualizer,
    ) -> None:
        self.models = models
        self.visualizer = visualizer
        self.converters = []
        for model in self.models:
            self.converters.append(create_output_converter(model.task_type, model.labels))

    # pylint: disable=too-many-locals
    def single_run(self, input_image: np.ndarray) -> AnnotationSceneEntity:
        """Inference for single image."""
        current_objects = [(input_image, Annotation(Rectangle(0, 0, 1, 1), labels=[]))]
        result_scene = AnnotationSceneEntity([], AnnotationSceneKind.PREDICTION)
        for index, model in enumerate(self.models):
            new_objects = []
            for item, parent_annotation in current_objects:
                predictions, frame_meta = model.core_model(item)
                annotation_scene = self.converters[index].convert_to_annotation(predictions, frame_meta)
                for annotation in annotation_scene.annotations:
                    new_item, item_annotation = self.crop(item, parent_annotation, annotation)
                    new_objects.append((new_item, item_annotation))
                    if model.task_type.is_global:
                        for label in item_annotation.get_labels():
                            parent_annotation.append_label(label)
                    else:
                        result_scene.append_annotation(item_annotation)
            current_objects = new_objects
        return result_scene

    @staticmethod
    def crop(
        item: np.ndarray, parent_annotation: Annotation, item_annotation: Annotation
    ) -> Tuple[np.ndarray, Annotation]:
        """Crop operation between chain stages."""
        new_item = ShapeFactory.shape_as_rectangle(item_annotation.shape).crop_numpy_array(item)
        item_annotation.shape = item_annotation.shape.normalize_wrt_roi_shape(
            ShapeFactory.shape_as_rectangle(parent_annotation.shape)
        )
        return new_item, item_annotation

    def run(self, input_stream: Union[int, str], loop: bool = False) -> None:
        """Run demo using input stream (image, video stream, camera)."""
        streamer = get_streamer(input_stream, loop)

        for frame in streamer:
            # getting result for single image
            annotation_scene = self.single_run(frame)
            output = self.visualizer.draw(frame, annotation_scene, {})
            self.visualizer.show(output)
            if self.visualizer.is_quit():
                break



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from abc import ABCMeta, abstractmethod

import torch
import torch.nn as nn
from mmseg.core import build_pixel_sampler
from scipy.special import erfinv

from otx.mpa.modules.models.builder import build_scalar_scheduler


class BaseWeightedLoss(nn.Module, metaclass=ABCMeta):
    """Base class for loss.

    All subclass should overwrite the ``_forward()`` method which returns the
    normal loss without loss weights.

    Args:
        loss_weight (float or dict): Factor scalar multiplied on the loss.
            Default: 1.0.
    """

    def __init__(
        self,
        reduction="mean",
        loss_weight=1.0,
        ignore_index=255,
        sampler=None,
        loss_jitter_prob=None,
        loss_jitter_momentum=0.1,
        **kwargs
    ):
        super().__init__()

        self.reduction = reduction
        self.ignore_index = ignore_index

        self.sampler = None
        if sampler is not None:
            self.sampler = build_pixel_sampler(sampler, ignore_index=ignore_index)

        self._smooth_loss = None
        self._jitter_sigma_factor = None
        self._loss_jitter_momentum = loss_jitter_momentum
        assert 0.0 < self._loss_jitter_momentum < 1.0
        if loss_jitter_prob is not None:
            assert 0.0 < loss_jitter_prob < 1.0
            self._jitter_sigma_factor = 1.0 / ((2.0**0.5) * erfinv(1.0 - 2.0 * loss_jitter_prob))

        self._loss_weight_scheduler = build_scalar_scheduler(loss_weight, default_value=1.0)

        self._iter = 0
        self._last_loss_weight = 0
        self._epoch_size = 1

    def set_step_params(self, init_iter, epoch_size):
        assert init_iter >= 0
        assert epoch_size > 0

        self._iter = init_iter
        self._epoch_size = epoch_size

    @property
    def with_loss_jitter(self):
        return self._jitter_sigma_factor is not None

    @property
    def iter(self):
        return self._iter

    @property
    def epoch_size(self):
        return self._epoch_size

    @property
    def last_loss_weight(self):
        return self._last_loss_weight

    @abstractmethod
    def _forward(self, *args, **kwargs):
        pass

    def forward(self, *args, **kwargs):
        """Defines the computation performed at every call.

        Args:
            *args: The positional arguments for the corresponding
                loss.
            **kwargs: The keyword arguments for the corresponding
                loss.

        Returns:
            torch.Tensor: The calculated loss.
        """

        loss, meta = self._forward(*args, **kwargs)
        # make sure meta data are tensor as well for aggregation
        # when parsing loss in sgementator
        for k, v in meta.items():
            meta[k] = torch.tensor(v, dtype=loss.dtype, device=loss.device)

        if self.with_loss_jitter and loss.numel() == 1:
            if self._smooth_loss is None:
                self._smooth_loss = loss.item()
            else:
                self._smooth_loss = (
                    1.0 - self._loss_jitter_momentum
                ) * self._smooth_loss + self._loss_jitter_momentum * loss.item()

            jitter_sigma = self._jitter_sigma_factor * abs(self._smooth_loss)
            jitter_point = torch.normal(0.0, jitter_sigma, [], device=loss.device, dtype=loss.dtype)
            loss = (loss - jitter_point).abs() + jitter_point

        self._last_loss_weight = self._loss_weight_scheduler(self.iter, self.epoch_size)
        out_loss = self._last_loss_weight * loss

        self._iter += 1

        return out_loss, meta



"""Demo based on ModelAPI."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import sys
from argparse import SUPPRESS, ArgumentParser
from pathlib import Path

# pylint: disable=no-name-in-module, import-error
from otx.api.usecases.exportable_code.demo.demo_package import (
    AsyncExecutor,
    ChainExecutor,
    ModelContainer,
    SyncExecutor,
    create_visualizer,
)


def build_argparser():
    """Parses command line arguments."""
    parser = ArgumentParser(add_help=False)
    args = parser.add_argument_group("Options")
    args.add_argument(
        "-h",
        "--help",
        action="help",
        default=SUPPRESS,
        help="Show this help message and exit.",
    )
    args.add_argument(
        "-i",
        "--input",
        required=True,
        help="Required. An input to process. The input must be a single image, "
        "a folder of images, video file or camera id.",
    )
    args.add_argument(
        "-m",
        "--models",
        help="Required. Path to directory with trained model and configuration file. "
        "If you provide several models you will start the task chain pipeline with "
        "the provided models in the order in which they were specified.",
        nargs="+",
        required=True,
        type=Path,
    )
    args.add_argument(
        "-it",
        "--inference_type",
        help="Optional. Type of inference for single model.",
        choices=["sync", "async"],
        default="sync",
        type=str,
    )
    args.add_argument(
        "-l",
        "--loop",
        help="Optional. Enable reading the input in a loop.",
        default=False,
        action="store_true",
    )
    args.add_argument(
        "--no_show",
        help="Optional. Disables showing inference results on UI.",
        default=False,
        action="store_true",
    )
    args.add_argument(
        "-d",
        "--device",
        help="Optional. Device to infer the model.",
        choices=["CPU", "GPU"],
        default="CPU",
        type=str,
    )

    return parser


EXECUTORS = {
    "sync": SyncExecutor,
    "async": AsyncExecutor,
    "chain": ChainExecutor,
}


def get_inferencer_class(type_inference, models):
    """Return class for inference of models."""
    if len(models) > 1:
        type_inference = "chain"
        print("You started the task chain pipeline with the provided models in the order in which they were specified")
    return EXECUTORS[type_inference]


def main():
    """Main function that is used to run demo."""
    args = build_argparser().parse_args()
    # create models
    models = []
    for model_dir in args.models:
        model = ModelContainer(model_dir, device=args.device)
        models.append(model)

    inferencer = get_inferencer_class(args.inference_type, models)

    # create visualizer
    visualizer = create_visualizer(models[-1].task_type, no_show=args.no_show)

    if len(models) == 1:
        models = models[0]

    # create inferencer and run
    demo = inferencer(models, visualizer)
    demo.run(args.input, args.loop and not args.no_show)


if __name__ == "__main__":
    sys.exit(main() or 0)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
import torch.nn as nn
from mmcls.models.builder import HEADS
from mmcls.models.heads import MultiLabelClsHead
from mmcv.cnn import build_activation_layer, constant_init, normal_init

from .custom_multi_label_linear_cls_head import AnglularLinear


@HEADS.register_module()
class CustomMultiLabelNonLinearClsHead(MultiLabelClsHead):
    """Non-linear classification head for multilabel task.
    Args:
        num_classes (int): Number of categories.
        in_channels (int): Number of channels in the input feature map.
        loss (dict): Config of classification loss.
        scale (float): positive scale parameter.
        init_cfg (dict | optional): The extra init config of layers.
            Defaults to use dict(type='Normal', layer='Linear', std=0.01).
        normalized (bool): Normalize input features and weights in the last linar layer.
    """

    def __init__(
        self,
        num_classes,
        in_channels,
        hid_channels=1280,
        act_cfg=dict(type="ReLU"),
        scale=1.0,
        loss=dict(type="CrossEntropyLoss", use_sigmoid=True, reduction="mean", loss_weight=1.0),
        dropout=False,
        normalized=False,
    ):

        super(CustomMultiLabelNonLinearClsHead, self).__init__(loss=loss)

        self.in_channels = in_channels
        self.num_classes = num_classes
        self.hid_channels = hid_channels
        self.dropout = dropout
        self.normalized = normalized
        self.scale = scale

        if self.num_classes <= 0:
            raise ValueError(f"num_classes={num_classes} must be a positive integer")

        self._init_layers(act_cfg)

    def _init_layers(self, act_cfg):
        modules = [
            nn.Linear(self.in_channels, self.hid_channels),
            nn.BatchNorm1d(self.hid_channels),
            build_activation_layer(act_cfg),
        ]
        if self.dropout:
            modules.append(nn.Dropout(p=0.2))
        if self.normalized:
            modules.append(AnglularLinear(self.hid_channels, self.num_classes))
        else:
            modules.append(nn.Linear(self.hid_channels, self.num_classes))

        self.classifier = nn.Sequential(*modules)

    def init_weights(self):
        for m in self.classifier:
            if isinstance(m, nn.Linear):
                normal_init(m, mean=0, std=0.01, bias=0)
            elif isinstance(m, nn.BatchNorm1d):
                constant_init(m, 1)

    def loss(self, cls_score, gt_label, valid_label_mask=None):
        gt_label = gt_label.type_as(cls_score)
        num_samples = len(cls_score)
        losses = dict()

        # map difficult examples to positive ones
        _gt_label = torch.abs(gt_label)
        # compute loss
        loss = self.compute_loss(
            cls_score,
            _gt_label,
            valid_label_mask=valid_label_mask,
            avg_factor=num_samples,
        )
        losses["loss"] = loss / self.scale
        return losses

    def forward_train(self, x, gt_label, **kwargs):
        img_metas = kwargs.get("img_metas", False)
        gt_label = gt_label.type_as(x)
        cls_score = self.classifier(x) * self.scale

        valid_batch_mask = gt_label >= 0
        gt_label = gt_label[
            valid_batch_mask,
        ].view(gt_label.shape[0], -1)
        cls_score = cls_score[
            valid_batch_mask,
        ].view(cls_score.shape[0], -1)
        if img_metas:
            valid_label_mask = self.get_valid_label_mask(img_metas=img_metas)
            valid_label_mask = valid_label_mask.to(x.device)
            valid_label_mask = valid_label_mask[
                valid_batch_mask,
            ].view(valid_label_mask.shape[0], -1)
            losses = self.loss(cls_score, gt_label, valid_label_mask=valid_label_mask)
        else:
            losses = self.loss(cls_score, gt_label)
        return losses

    def simple_test(self, img):
        """Test without augmentation."""
        cls_score = self.classifier(img) * self.scale
        if isinstance(cls_score, list):
            cls_score = sum(cls_score) / float(len(cls_score))
        if torch.onnx.is_in_onnx_export():
            return cls_score
        pred = torch.sigmoid(cls_score) if cls_score is not None else None
        pred = list(pred.detach().cpu().numpy())
        return pred

    def get_valid_label_mask(self, img_metas):
        valid_label_mask = []
        for i, meta in enumerate(img_metas):
            mask = torch.Tensor([1 for _ in range(self.num_classes)])
            if "ignored_labels" in meta and meta["ignored_labels"]:
                mask[meta["ignored_labels"]] = 0
            valid_label_mask.append(mask)
        valid_label_mask = torch.stack(valid_label_mask, dim=0)
        return valid_label_mask



"""This module contains the interface class for tasks."""


# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc
from typing import Dict

import numpy as np

from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import InferenceParameters


class IInferenceTask(metaclass=abc.ABCMeta):
    """A base interface class for a task."""

    @abc.abstractmethod
    def infer(
        self,
        dataset: DatasetEntity,
        inference_parameters: InferenceParameters,
    ) -> DatasetEntity:
        """This is the method that is called upon inference.

        This happens when the user wants to analyse a sample
        or multiple samples need to be analysed.

        Args:
            dataset: The input dataset to perform the analysis on.
            inference_parameters: The parameters to use for the
                analysis.

        Returns:
            The results of the analysis.
        """
        raise NotImplementedError


class IRawInference(metaclass=abc.ABCMeta):
    """A base interface class for raw inference tasks."""

    @abc.abstractmethod
    def raw_infer(
        self,
        input_tensors: Dict[str, np.ndarray],
        output_tensors: Dict[str, np.ndarray],
    ):
        """This is the method that is called to run a neural network over a set of tensors.

        This method takes as input/output the tensors which are directly fed to the neural network,
        and does not include any additional pre- and post-processing of the inputs and outputs.

        Args:
            input_tensors: Dictionary containing the input tensors.
            output_tensors: Dictionary to be filled by the task with the
                output tensors.
        """
        raise NotImplementedError



"""Utilities for OpenVINO telemetry."""
# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=broad-exception-caught
import json

import openvino_telemetry as tm

from otx import __version__

__TM_CATEGORY_OTX = "otx"
__TM_MEASUREMENT_ID = "UA-17808594-29"
# __TM_MEASUREMENT_ID_FOR_TESTING = "UA-254359572-1"
# __TM_MEASUREMENT_ID = __TM_MEASUREMENT_ID_FOR_TESTING

__TM_ACTION_VERSION = "version"
__TM_ACTION_CMD_SUCCESS = "success"
__TM_ACTION_CMD_FAILURE = "failure"
__TM_ACTION_CMD_EXCEPTION = "exception"
__TM_ACTION_ERROR = "error"


def init_telemetry_session():
    """Init session."""
    telemetry = tm.Telemetry(app_name=__TM_CATEGORY_OTX, app_version=str(__version__), tid=__TM_MEASUREMENT_ID)
    telemetry.start_session(__TM_CATEGORY_OTX)
    send_version(telemetry)

    return telemetry


def close_telemetry_session(telemetry):
    """Close session."""
    if not isinstance(telemetry, tm.Telemetry):
        raise RuntimeError(f"Invalid argument. required {type(tm.Telemetry)} but passed {type(telemetry)}")
    telemetry.end_session(__TM_CATEGORY_OTX)
    telemetry.force_shutdown(1.0)


def send_version(telemetry):
    """Send application version."""
    if not isinstance(telemetry, tm.Telemetry):
        raise RuntimeError(f"Invalid argument. required {type(tm.Telemetry)} but passed {type(telemetry)}")
    __send_event(telemetry, __TM_ACTION_VERSION, str(__version__))


def send_cmd_results(telemetry, cmd, results):
    """Send cli telemetry data."""
    if not isinstance(telemetry, tm.Telemetry):
        raise RuntimeError(f"Invalid argument. required {type(tm.Telemetry)} but passed {type(telemetry)}")
    action = __TM_ACTION_ERROR

    if not isinstance(results, dict):
        raise RuntimeError(f"Invalid argument. required {dict} but passed {type(results)}")

    retcode = results.pop("retcode", None)
    if retcode is not None:
        label = dict(cmd=cmd, **results)
        if retcode >= 0:
            action = __TM_ACTION_CMD_FAILURE
            if retcode == 0:
                action = __TM_ACTION_CMD_SUCCESS
        else:
            action = __TM_ACTION_CMD_EXCEPTION
            label = dict(cmd=cmd, **results)

    if action == __TM_ACTION_ERROR:
        __send_error(telemetry, f"Invalid results for sending cmd result: {results}")
    else:
        __send_event(telemetry, action, label)


def __send_event(telemetry, action, label, **kwargs):
    """Wrapper of the openvino-telemetry.send_event()."""
    try:
        telemetry.send_event(__TM_CATEGORY_OTX, action, json.dumps(label), **kwargs)
    except Exception as error:
        print(f"An error while calling otm.send_event(): \n{repr(error)}")


def __send_error(telemetry, err_msg, **kwargs):
    """Wrapper of the openvino-telemetry.send_error()."""
    try:
        telemetry.send_error(__TM_CATEGORY_OTX, err_msg, **kwargs)
    except Exception as error:
        print(f"An error while calling otm.send_error(): \n{repr(error)}")



"""This module contains the different ui rules elements, Rule and UIRules.

They are used to define rules for disabling configuration parameters in the ui, conditional on the value of other
parameters.
"""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from typing import Callable, List, Optional, Union

from attr import asdict, attrib, attrs, setters

from otx.api.configuration.elements.utils import attr_enum_to_str_serializer
from otx.api.configuration.enums.config_element_type import ConfigElementType

from .types import Action, Operator
from .utils import attr_convert_action, attr_convert_operator

ALLOWED_RULE_VALUE_TYPES = Union[int, str, float, bool]  # pylint: disable=invalid-name


@attrs(auto_attribs=True)
class Rule:
    """This class represents a `operator` applied to the `value` of the configurable parameter `parameter`.

    The parameter for which the rule should be evaluated is identified by name, or by a list of names representing the
    attribute path to the parameter in case of a nested configuration.
    """

    parameter: Union[str, List[str]]
    value: ALLOWED_RULE_VALUE_TYPES
    operator: Operator = attrib(default=Operator.EQUAL_TO, converter=attr_convert_operator)
    type: ConfigElementType = attrib(default=ConfigElementType.RULE, on_setattr=setters.frozen)

    def to_dict(self, enum_to_str: bool = True) -> dict:
        """Method to serialize a Rule instance to its dictionary representation.

        Args:
            enum_to_str (bool): Set to True to convert Enum instances to their string representation

        Returns:
            dict: dictionary representation of the Rule object for which this method is called
        """
        if enum_to_str:
            serializer: Optional[Callable] = attr_enum_to_str_serializer
        else:
            serializer = None
        return asdict(self, value_serializer=serializer)


@attrs(auto_attribs=True)
class UIRules:
    """This class allows the combination of ExposureRules using boolean logic.

    The class can be set as an attribute of a configurable parameter. If the `rules`
    (combined according to the `operator`) evaluate to True, the corresponding`action` will be taken in the UI.

    If UIRules are nested, only the `action` of the outermost UIRule will be considered.
    """

    rules: List[Union[Rule, UIRules]] = attrib(kw_only=True)
    operator: Operator = attrib(default=Operator.AND, converter=attr_convert_operator)
    action: Action = attrib(default=Action.DISABLE_EDITING, converter=attr_convert_action)
    type: ConfigElementType = attrib(default=ConfigElementType.UI_RULES, on_setattr=setters.frozen)

    def add_rule(self, rule: Union[Rule, UIRules]):
        """Adds rule."""
        self.rules.append(rule)

    def to_dict(self, enum_to_str: bool = True) -> dict:
        """Method to serialize an UIRules instance to its dictionary representation.

        Applies recursion to convert nested rules, if applicable.

        Args:
            enum_to_str: Set to True to convert Enum instances to their
                string representation

        Returns:
            dictionary representation of the UIRules object for which
            this method is called
        """
        if enum_to_str:
            serializer: Optional[Callable] = attr_enum_to_str_serializer
        else:
            serializer = None
        rules_list = []
        for rule in self.rules:
            rules_list.append(rule.to_dict(enum_to_str))
        dictionary_representation = asdict(self, value_serializer=serializer)
        dictionary_representation.update({"rules": rules_list})
        return dictionary_representation


@attrs
class NullUIRules(UIRules):
    """This class represents an empty, unset UIRules element."""

    rules: List[Union[Rule, UIRules]] = attrib(factory=list)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
from mmcls.models.builder import CLASSIFIERS
from mmcls.models.classifiers.image import ImageClassifier
from torch.nn.functional import sigmoid, softmax


@CLASSIFIERS.register_module()
class SupConClassifier(ImageClassifier):
    def __init__(self, backbone, neck=None, head=None, pretrained=None, **kwargs):
        self.multilabel = kwargs.pop("multilabel", False)
        self.hierarchical = kwargs.pop("hierarchical", False)
        super().__init__(backbone, neck=neck, head=head, pretrained=pretrained, **kwargs)

    def forward_train(self, img, gt_label, **kwargs):
        # concatenate the different image views along the batch size
        if len(img.shape) == 5:
            img = torch.cat([img[:, d, :, :, :] for d in range(img.shape[1])], dim=0)
        x = self.extract_feat(img)
        losses = dict()
        if self.multilabel or self.hierarchical:
            loss = self.head.forward_train(x, gt_label, **kwargs)
        else:
            gt_label = gt_label.squeeze(dim=1)
            loss = self.head.forward_train(x, gt_label)
        losses.update(loss)
        return losses



"""This module contains a default set of configurable parameters for a model."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from attr import attrib, attrs

from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.configuration.elements import (
    ParameterGroup,
    add_parameter_group,
    boolean_attribute,
    configurable_float,
    configurable_integer,
    string_attribute,
)
from otx.api.configuration.enums.model_lifecycle import ModelLifecycle


@attrs
class DefaultModelParameters(ConfigurableParameters):
    """Configuration element representing a the default set of hyper parameters for a model.

    Attributes:
        header (str): Name of parameter group
        description (str): User friendly string describing what the ModelConfig represents, that will be displayed in
            the UI.
    """

    header: str = attrib(default="Default model hyper parameters")
    description: str = attrib(default="Default model hyper parameter section description", kw_only=True)

    @attrs
    class _LearningParameters(ParameterGroup):
        # Set defaults for the learning parameters. Learning parameters consist of at
        # least batch_size, epochs and learning_rate. These correspond to the 'basic'
        # hyper parameters in model template
        header = string_attribute("Learning Parameters")
        description = string_attribute("Parameters to control basic training behavior.")
        visible_in_ui = boolean_attribute(True)

        batch_size = configurable_integer(
            header="Batch size",
            description="The number of training samples seen in each "
            "iteration of training. Setting this higher will "
            "make the training more stable, but will require "
            "more memory. Setting this lower will make the "
            "training less stable, but will require less "
            "memory.",
            warning="Increasing this value may cause the system to use "
            "more memory than available, potentially causing out "
            "of memory errors, please update with caution.",
            min_value=1,
            max_value=1000,
            default_value=4,
            affects_outcome_of=ModelLifecycle.TRAINING,
        )
        epochs = configurable_integer(
            header="Number of epochs",
            default_value=10,
            min_value=1,
            max_value=10000,
            description="Increasing this value causes the results to be more "
            "robust but training time will be longer.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )
        learning_rate = configurable_float(
            header="Learning rate",
            default_value=1e-3,
            min_value=1e-30,
            max_value=1e10,
            description="Increasing this value will speed up training " "convergence but might make it unstable.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

    learning_parameters = add_parameter_group(_LearningParameters)



"""Classification Dataset Adapter."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=invalid-name, too-many-locals, no-member
from typing import List, Union

from datumaro.components.annotation import AnnotationType, LabelCategories

from otx.api.entities.annotation import Annotation
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.image import Image
from otx.api.entities.label import LabelEntity
from otx.api.entities.label_schema import LabelGroup, LabelGroupType, LabelSchemaEntity
from otx.core.data.adapter.base_dataset_adapter import BaseDatasetAdapter


class ClassificationDatasetAdapter(BaseDatasetAdapter):
    """Classification adapter inherited from BaseDatasetAdapter.

    It converts DatumaroDataset -> DatasetEntity
    for multi-class, multi-label, and hierarchical-label classification tasks
    """

    def get_otx_dataset(self) -> DatasetEntity:
        """Convert DatumaroDataset to DatasetEntity for Classification."""
        # Prepare label information
        label_information = self._prepare_label_information(self.dataset)
        self.category_items = label_information["category_items"]
        self.label_groups = label_information["label_groups"]
        self.label_entities = label_information["label_entities"]

        # Generate label schema
        self.label_schema = self._generate_classification_label_schema(self.label_groups, self.label_entities)

        # Set the DatasetItemEntity
        dataset_items: List[DatasetItemEntity] = []
        for subset, subset_data in self.dataset.items():
            for _, datumaro_items in subset_data.subsets().items():
                for datumaro_item in datumaro_items:
                    image = Image(file_path=datumaro_item.media.path)
                    shapes: List[Annotation] = []
                    for ann in datumaro_item.annotations:
                        if ann.type == AnnotationType.label:
                            shapes.append(self._get_label_entity(ann))

                    dataset_item = DatasetItemEntity(image, self._get_ann_scene_entity(shapes), subset=subset)
                    dataset_items.append(dataset_item)

        return DatasetEntity(items=dataset_items)

    def get_label_schema(self) -> LabelSchemaEntity:
        """Get Label Schema."""
        return self._generate_classification_label_schema(self.label_groups, self.label_entities)

    def _generate_classification_label_schema(
        self, label_groups: List[LabelCategories.LabelGroup], label_entities: List[LabelEntity]
    ) -> LabelSchemaEntity:
        """Generate LabelSchema for Classification."""
        label_schema = LabelSchemaEntity()

        if len(label_groups) > 0:
            for label_group in label_groups:
                group_label_entity_list = []
                for label in label_group.labels:
                    label_entity = [le for le in label_entities if le.name == label]
                    group_label_entity_list.append(label_entity[0])

                label_schema.add_group(
                    LabelGroup(
                        name=label_group.name, labels=group_label_entity_list, group_type=LabelGroupType.EXCLUSIVE
                    )
                )
            label_schema.add_group(self._generate_empty_label_entity())
        else:
            label_schema = self._generate_default_label_schema(label_entities)

        return label_schema

    def _select_data_type(self, data_candidates: Union[list, str]) -> str:
        return "imagenet" if "imagenet" in data_candidates else data_candidates[0]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import importlib


def load_ext(path, funcs):
    ext = importlib.import_module(path)
    for fun in funcs:
        assert hasattr(ext, fun), f"{fun} miss in module {path}"

    return ext



"""Collection of utils for dataset in Segmentation Task."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import json
import os
from typing import List, Optional

import cv2
import numpy as np
import tqdm
from mmseg.datasets.custom import CustomDataset
from skimage.segmentation import felzenszwalb

from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.id import ID
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.subset import Subset
from otx.api.utils.argument_checks import (
    DirectoryPathCheck,
    JsonFilePathCheck,
    OptionalDirectoryPathCheck,
    check_input_parameters_type,
)
from otx.mpa.utils.logger import get_logger

logger = get_logger()

# pylint: disable=too-many-locals


@check_input_parameters_type({"annot_path": JsonFilePathCheck})
def get_classes_from_annotation(annot_path):
    """Getter function of classes from annotation."""
    with open(annot_path, encoding="UTF-8") as input_stream:
        content = json.load(input_stream)
        labels_map = content["labels_map"]

        categories = [(v["name"], v["id"]) for v in sorted(labels_map, key=lambda tup: int(tup["id"]))]

    return categories


@check_input_parameters_type({"value": OptionalDirectoryPathCheck})
def abs_path_if_valid(value):
    """Valid function of abs_path."""
    if value:
        return os.path.abspath(value)
    return None


@check_input_parameters_type()
def create_annotation_from_hard_seg_map(hard_seg_map: np.ndarray, labels: List[LabelEntity]):
    """Creation function from hard seg_map."""
    height, width = hard_seg_map.shape[:2]
    unique_labels = np.unique(hard_seg_map)

    annotations: List[Annotation] = []
    for label_id in unique_labels:
        label_id_entity = ID(f"{label_id:08}")
        matches = [label for label in labels if label.id == label_id_entity]
        if len(matches) == 0:
            continue

        assert len(matches) == 1
        label = matches[0]

        label_mask = hard_seg_map == label_id
        label_index_map = label_mask.astype(np.uint8)

        contours, hierarchies = cv2.findContours(label_index_map, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE)

        if hierarchies is None:
            continue

        for contour, hierarchy in zip(contours, hierarchies[0]):
            if hierarchy[3] != -1:
                continue

            contour = list(contour)
            if len(contour) <= 2:
                continue

            points = [Point(x=point[0][0] / width, y=point[0][1] / height) for point in contour]

            annotations.append(
                Annotation(
                    Polygon(points=points),
                    labels=[ScoredLabel(label)],
                    id=ID(f"{label_id:08}"),
                )
            )

    return annotations


@check_input_parameters_type({"ann_dir": OptionalDirectoryPathCheck})
def load_labels_from_annotation(ann_dir):
    """Load labels function from annotation."""
    if ann_dir is None:
        return []

    labels_map_path = os.path.join(ann_dir, "meta.json")
    labels = get_classes_from_annotation(labels_map_path)

    return labels


@check_input_parameters_type()
def add_labels(cur_labels: List[LabelEntity], new_labels: List[tuple]):
    """Add labels function."""
    for label_name, label_id in new_labels:
        matching_labels = [label for label in cur_labels if label.name == label_name]
        if len(matching_labels) > 1:
            raise ValueError("Found multiple matching labels")
        if len(matching_labels) == 0:
            label_id = label_id if label_id is not None else len(cur_labels)
            label = LabelEntity(name=label_name, domain=Domain.SEGMENTATION, id=ID(f"{label_id:08}"))
            cur_labels.append(label)


@check_input_parameters_type()
def check_labels(cur_labels: List[LabelEntity], new_labels: List[tuple]):
    """Check labels function."""
    cur_names = {label.name for label in cur_labels}
    new_names = {label[0] for label in new_labels}
    if cur_names != new_names:
        raise ValueError("Class names don't match from file to file")


@check_input_parameters_type()
def get_extended_label_names(labels: List[LabelEntity]):
    """Getter function of extended label names."""
    target_labels = [v.name for v in sorted(labels, key=lambda x: x.id)]
    all_labels = ["background"] + target_labels
    return all_labels


@check_input_parameters_type()
def create_pseudo_masks(ann_file_path: str, data_root_dir: str, mode="FH"):
    """Create pseudo masks for Self-SL using DetCon."""
    if not os.path.isdir(ann_file_path):
        logger.info(
            (
                f"Creating pseudo masks with mode={mode} is required. "
                f"It may take some time. Once this process has been performed, "
                f"there is no need to proceed again with "
                f"ann_file_path={ann_file_path} and data_root_dir={data_root_dir}."
            )
        )
        os.makedirs(ann_file_path, exist_ok=False)
        img_list = os.listdir(data_root_dir)
        total_labels = []
        # create pseudo masks
        for path in tqdm.tqdm(img_list, total=len(img_list)):
            save_path = path.replace(".jpg", ".png") if path.endswith(".jpg") else path
            img = cv2.imread(os.path.join(data_root_dir, path))[..., ::-1]
            if mode == "FH":
                pseudo_mask = felzenszwalb(img, scale=1000, min_size=1000)
            else:
                raise ValueError(
                    (f"{mode} is not supported to create pseudo masks for DetCon." 'Choose one of ["FH"].')
                )
            cv2.imwrite(os.path.join(ann_file_path, save_path), pseudo_mask.astype(np.uint8))

            # get labels to create meta.json
            labels = np.unique(pseudo_mask)
            for label in labels:
                if label not in total_labels:
                    total_labels.append(label)

        # create meta.json
        # TODO (sungchul): to be updated as max(total_labels) -> max(total_labels)+1
        # Currently, background class is automatically added in the backend.
        # Considering background class, labels_map in meta.json should have one less than the number of labels.
        # If we don't need to consider background class, it will be updated to consider all labels.
        meta = {"labels_map": [{"name": f"target{i+1}", "id": i + 1} for i in range(max(total_labels))]}
        with open(os.path.join(ann_file_path, "meta.json"), "w", encoding="UTF-8") as f:
            json.dump(meta, f, indent=4)


@check_input_parameters_type({"ann_file_path": DirectoryPathCheck, "data_root_dir": DirectoryPathCheck})
def load_dataset_items(
    ann_file_path: str,
    data_root_dir: str,
    subset: Subset = Subset.NONE,
    labels_list: Optional[List[LabelEntity]] = None,
):
    """Load dataset items."""
    if "detcon" in ann_file_path:  # TODO (sungchul): deterministic condition
        create_pseudo_masks(ann_file_path, data_root_dir)

    ann_dir = abs_path_if_valid(ann_file_path)
    img_dir = abs_path_if_valid(data_root_dir)

    annot_labels = load_labels_from_annotation(ann_dir)

    if labels_list is None:
        labels_list = []
    if len(labels_list) == 0:
        add_labels(labels_list, annot_labels)
    else:
        check_labels(labels_list, annot_labels)

    test_mode = subset in {Subset.VALIDATION, Subset.TESTING}
    pipeline = [dict(type="LoadAnnotations")]

    dataset = CustomDataset(
        img_dir=img_dir,
        ann_dir=ann_dir,
        pipeline=pipeline,
        classes=get_extended_label_names(labels_list),
        test_mode=test_mode,
    )
    dataset.test_mode = False

    dataset_items = []
    for item in dataset:
        annotations = create_annotation_from_hard_seg_map(hard_seg_map=item["gt_semantic_seg"], labels=labels_list)
        filename = os.path.join(item["img_prefix"], item["img_info"]["filename"])
        image = Image(file_path=filename)
        annotation_scene = AnnotationSceneEntity(kind=AnnotationSceneKind.ANNOTATION, annotations=annotations)
        dataset_items.append(DatasetItemEntity(media=image, annotation_scene=annotation_scene, subset=subset))

    return dataset_items



"""Inference Task of OTX Detection."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import os
from typing import Iterable, List, Optional, Tuple

import cv2
import numpy as np
import pycocotools.mask as mask_util
from mmcv.utils import ConfigDict

from otx.algorithms.common.adapters.mmcv.utils import (
    get_configs_by_keys,
    patch_data_pipeline,
    patch_default_config,
    patch_runner,
)
from otx.algorithms.common.configs.training_base import TrainType
from otx.algorithms.common.tasks.training_base import BaseTask
from otx.algorithms.common.utils.callback import InferenceProgressCallback
from otx.algorithms.detection.adapters.mmdet.utils import (
    patch_datasets,
    patch_evaluation,
)
from otx.algorithms.detection.adapters.mmdet.utils.builder import build_detector
from otx.algorithms.detection.adapters.mmdet.utils.config_utils import (
    cluster_anchors,
    should_cluster_anchors,
)
from otx.algorithms.detection.configs.base import DetectionConfig
from otx.api.configuration.helper.utils import config_to_bytes
from otx.api.entities.annotation import Annotation
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.model import (
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
    ModelPrecision,
)
from otx.api.entities.model_template import TaskType
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.entities.tensor import TensorEntity
from otx.api.entities.train_parameters import default_progress_callback
from otx.api.serialization.label_mapper import label_schema_to_bytes
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.tasks.interfaces.evaluate_interface import IEvaluationTask
from otx.api.usecases.tasks.interfaces.explain_interface import IExplainTask
from otx.api.usecases.tasks.interfaces.export_interface import ExportType, IExportTask
from otx.api.usecases.tasks.interfaces.inference_interface import IInferenceTask
from otx.api.usecases.tasks.interfaces.unload_interface import IUnload
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)
from otx.api.utils.dataset_utils import add_saliency_maps_to_dataset_item
from otx.mpa import MPAConstants
from otx.mpa.utils.config_utils import MPAConfig
from otx.mpa.utils.logger import get_logger

logger = get_logger()

RECIPE_TRAIN_TYPE = {
    TrainType.SEMISUPERVISED: "semisl.py",
    TrainType.INCREMENTAL: "incremental.py",
}


# pylint: disable=too-many-locals, too-many-instance-attributes
class DetectionInferenceTask(BaseTask, IInferenceTask, IExportTask, IEvaluationTask, IExplainTask, IUnload):
    """Inference Task Implementation of OTX Detection."""

    @check_input_parameters_type()
    def __init__(self, task_environment: TaskEnvironment, **kwargs):
        # self._should_stop = False
        super().__init__(DetectionConfig, task_environment, **kwargs)
        self.template_dir = os.path.abspath(os.path.dirname(self.template_file_path))

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def infer(
        self,
        dataset: DatasetEntity,
        inference_parameters: Optional[InferenceParameters] = None,
    ) -> DatasetEntity:
        """Main infer function of OTX Detection."""
        logger.info("infer()")

        update_progress_callback = default_progress_callback
        process_saliency_maps = False
        explain_predicted_classes = True
        if inference_parameters is not None:
            update_progress_callback = inference_parameters.update_progress  # type: ignore
            process_saliency_maps = inference_parameters.process_saliency_maps
            explain_predicted_classes = inference_parameters.explain_predicted_classes

        self._time_monitor = InferenceProgressCallback(len(dataset), update_progress_callback)
        # If confidence threshold is adaptive then up-to-date value should be stored in the model
        # and should not be changed during inference. Otherwise user-specified value should be taken.
        if not self._hyperparams.postprocessing.result_based_confidence_threshold:
            self.confidence_threshold = self._hyperparams.postprocessing.confidence_threshold
        logger.info(f"Confidence threshold {self.confidence_threshold}")

        prediction_results, _ = self._infer_detector(dataset, inference_parameters)
        self._add_predictions_to_dataset(
            prediction_results, dataset, self.confidence_threshold, process_saliency_maps, explain_predicted_classes
        )
        logger.info("Inference completed")
        return dataset

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def explain(
        self,
        dataset: DatasetEntity,
        explain_parameters: Optional[InferenceParameters] = None,
    ) -> DatasetEntity:
        """Main explain function of OTX Detection."""
        logger.info("explain()")

        update_progress_callback = default_progress_callback
        process_saliency_maps = False
        explain_predicted_classes = True
        if explain_parameters is not None:
            update_progress_callback = explain_parameters.update_progress  # type: ignore
            process_saliency_maps = explain_parameters.process_saliency_maps
            explain_predicted_classes = explain_parameters.explain_predicted_classes

        self._time_monitor = InferenceProgressCallback(len(dataset), update_progress_callback)
        detections, explain_results = self._explain_detector(dataset, explain_parameters)
        self._add_explanations_to_dataset(
            detections, explain_results, dataset, process_saliency_maps, explain_predicted_classes
        )
        logger.info("Explain completed")
        return dataset

    def _infer_detector(
        self,
        dataset: DatasetEntity,
        inference_parameters: Optional[InferenceParameters] = None,
    ) -> Tuple[Iterable, float]:
        """Inference wrapper.

        This method triggers the inference and returns `prediction_results` zipped with prediction results,
        feature vectors, and saliency maps. `metric` is returned as a float value if InferenceParameters.is_evaluation
        is set to true, otherwise, `None` is returned.

        Args:
            dataset (DatasetEntity): the validation or test dataset to be inferred with
            inference_parameters (Optional[InferenceParameters], optional): Option to run evaluation or not.
                If `InferenceParameters.is_evaluation=True` then metric is returned, otherwise, both metric and
                saliency maps are empty. Defaults to None.

        Returns:
            Tuple[Iterable, float]: Iterable prediction results for each sample and metric for on the given dataset
        """
        stage_module = "DetectionInferrer"
        self._data_cfg = self._init_test_data_cfg(dataset)

        dump_features = True
        dump_saliency_map = not inference_parameters.is_evaluation if inference_parameters else True

        results = self._run_task(
            stage_module,
            mode="train",
            dataset=dataset,
            eval=inference_parameters.is_evaluation if inference_parameters else False,
            dump_features=dump_features,
            dump_saliency_map=dump_saliency_map,
        )
        # TODO: InferenceProgressCallback register
        logger.debug(f"result of run_task {stage_module} module = {results}")
        output = results["outputs"]
        metric = output["metric"]
        predictions = output["detections"]
        assert len(output["detections"]) == len(output["feature_vectors"]) == len(output["saliency_maps"]), (
            "Number of elements should be the same, however, number of outputs are "
            f"{len(output['detections'])}, {len(output['feature_vectors'])}, and {len(output['saliency_maps'])}"
        )
        prediction_results = zip(predictions, output["feature_vectors"], output["saliency_maps"])
        return prediction_results, metric

    def _explain_detector(
        self,
        dataset: DatasetEntity,
        explain_parameters: Optional[InferenceParameters] = None,
    ) -> Tuple[List[List[np.array]], List[np.array]]:
        """Run explain stage and return detections and saliency maps."""

        stage_module = "DetectionExplainer"
        self._data_cfg = self._init_test_data_cfg(dataset)
        results = self._run_task(
            stage_module,
            mode="train",
            dataset=dataset,
            explainer=explain_parameters.explainer if explain_parameters else None,
        )
        detections = results["outputs"]["detections"]
        explain_results = results["outputs"]["saliency_maps"]
        return detections, explain_results

    @check_input_parameters_type()
    def evaluate(
        self,
        output_resultset: ResultSetEntity,
        evaluation_metric: Optional[str] = None,
    ):
        """Evaluate function of OTX Detection Task."""
        logger.info("called evaluate()")
        if evaluation_metric is not None:
            logger.warning(
                f"Requested to use {evaluation_metric} metric, " "but parameter is ignored. Use F-measure instead."
            )
        metric = MetricsHelper.compute_f_measure(output_resultset)
        logger.info(f"F-measure after evaluation: {metric.f_measure.value}")
        output_resultset.performance = metric.get_performance()
        logger.info("Evaluation completed")

    def unload(self):
        """Unload the task."""
        self.cleanup()

    @check_input_parameters_type()
    def export(self, export_type: ExportType, output_model: ModelEntity, dump_features: bool = False):
        """Export function of OTX Detection Task."""
        # copied from OTX inference_task.py
        logger.info("Exporting the model")
        if export_type != ExportType.OPENVINO:
            raise RuntimeError(f"not supported export type {export_type}")
        output_model.model_format = ModelFormat.OPENVINO
        output_model.optimization_type = ModelOptimizationType.MO

        stage_module = "DetectionExporter"
        results = self._run_task(
            stage_module,
            mode="train",
            export=True,
            dump_features=dump_features,
        )
        outputs = results.get("outputs")
        logger.debug(f"results of run_task = {outputs}")
        if outputs is None:
            raise RuntimeError(results.get("msg"))

        bin_file = outputs.get("bin")
        xml_file = outputs.get("xml")
        if xml_file is None or bin_file is None:
            raise RuntimeError("invalid status of exporting. bin and xml should not be None")
        with open(bin_file, "rb") as f:
            output_model.set_data("openvino.bin", f.read())
        with open(xml_file, "rb") as f:
            output_model.set_data("openvino.xml", f.read())
        output_model.set_data(
            "confidence_threshold",
            np.array([self.confidence_threshold], dtype=np.float32).tobytes(),
        )
        output_model.set_data("config.json", config_to_bytes(self._hyperparams))
        output_model.precision = [ModelPrecision.FP32]
        output_model.optimization_methods = self._optimization_methods
        output_model.set_data(
            "label_schema.json",
            label_schema_to_bytes(self._task_environment.label_schema),
        )
        logger.info("Exporting completed")

    def _init_recipe_hparam(self) -> dict:
        configs = super()._init_recipe_hparam()
        # Update tiling parameters if tiling is enabled
        if bool(self._hyperparams.tiling_parameters.enable_tiling):
            logger.info("Tiling Enabled")
            tiling_params = ConfigDict(
                tile_size=int(self._hyperparams.tiling_parameters.tile_size),
                overlap_ratio=float(self._hyperparams.tiling_parameters.tile_overlap),
                max_per_img=int(self._hyperparams.tiling_parameters.tile_max_number),
            )
            configs.update(
                ConfigDict(
                    data=ConfigDict(
                        train=tiling_params,
                        val=tiling_params,
                        test=tiling_params,
                    )
                )
            )
            configs.update(dict(evaluation=dict(iou_thr=[0.5])))

        configs["use_adaptive_interval"] = self._hyperparams.learning_parameters.use_adaptive_interval
        return configs

    def _init_recipe(self):
        logger.info("called _init_recipe()")

        recipe_root = os.path.join(MPAConstants.RECIPES_PATH, "stages/detection")
        if self._task_type.domain in {
            Domain.INSTANCE_SEGMENTATION,
            Domain.ROTATED_DETECTION,
        }:
            recipe_root = os.path.join(MPAConstants.RECIPES_PATH, "stages/instance-segmentation")

        logger.info(f"train type = {self._train_type}")

        # Incremental Learning Recipe is the default for Detection
        if self._train_type in RECIPE_TRAIN_TYPE:
            recipe = os.path.join(recipe_root, RECIPE_TRAIN_TYPE[self._train_type])
        else:
            raise NotImplementedError(f"Train type {self._train_type} is not implemented yet.")

        logger.info(f"train type = {self._train_type} - loading {recipe}")

        self._recipe_cfg = MPAConfig.fromfile(recipe)

        options_for_patch_datasets = {"type": "OTXDetDataset"}
        patch_default_config(self._recipe_cfg)
        patch_runner(self._recipe_cfg)
        patch_data_pipeline(self._recipe_cfg, self.data_pipeline_path)
        patch_datasets(
            self._recipe_cfg,
            self._task_type.domain,
            **options_for_patch_datasets,
        )  # for OTX compatibility
        patch_evaluation(self._recipe_cfg)  # for OTX compatibility
        logger.info(f"initialized recipe = {recipe}")

    def _init_model_cfg(self):
        model_cfg = MPAConfig.fromfile(os.path.join(self._model_dir, "model.py"))
        if len(self._anchors) != 0:
            self._update_anchors(model_cfg.model.bbox_head.anchor_generator, self._anchors)
        return model_cfg

    def _init_test_data_cfg(self, dataset: DatasetEntity):
        data_cfg = ConfigDict(
            data=ConfigDict(
                train=ConfigDict(
                    otx_dataset=None,
                    labels=self._labels,
                ),
                test=ConfigDict(
                    otx_dataset=dataset,
                    labels=self._labels,
                ),
            )
        )
        return data_cfg

    def _update_stage_module(self, stage_module):
        module_prefix = {TrainType.INCREMENTAL: "Incr", TrainType.SEMISUPERVISED: "SemiSL"}
        if self._train_type == TrainType.SEMISUPERVISED and stage_module == "DetectionExporter":
            stage_module = "SemiSLDetectionExporter"
        elif self._train_type in module_prefix and stage_module in [
            "DetectionTrainer",
            "DetectionInferrer",
        ]:
            stage_module = module_prefix[self._train_type] + stage_module
        return stage_module

    def _get_shapes(self, all_results, width, height, confidence_threshold):
        if self._task_type == TaskType.DETECTION:
            shapes = self._det_add_predictions_to_dataset(all_results, width, height, confidence_threshold)
        elif self._task_type in {
            TaskType.INSTANCE_SEGMENTATION,
            TaskType.ROTATED_DETECTION,
        }:
            shapes = self._ins_seg_add_predictions_to_dataset(all_results, width, height, confidence_threshold)
        else:
            raise RuntimeError(f"MPA results assignment not implemented for task: {self._task_type}")
        return shapes

    def _add_predictions_to_dataset(
        self,
        prediction_results,
        dataset,
        confidence_threshold=0.0,
        process_saliency_maps=False,
        explain_predicted_classes=True,
    ):
        """Loop over dataset again to assign predictions. Convert from MMDetection format to OTX format."""
        for dataset_item, (all_results, feature_vector, saliency_map) in zip(dataset, prediction_results):
            shapes = self._get_shapes(all_results, dataset_item.width, dataset_item.height, confidence_threshold)
            dataset_item.append_annotations(shapes)

            if feature_vector is not None:
                active_score = TensorEntity(name="representation_vector", numpy=feature_vector.reshape(-1))
                dataset_item.append_metadata_item(active_score, model=self._task_environment.model)

            if saliency_map is not None:
                labels = self._labels.copy()
                if saliency_map.shape[0] == len(labels) + 1:
                    # Include the background as the last category
                    labels.append(LabelEntity("background", Domain.DETECTION))

                predicted_scored_labels = []
                for shape in shapes:
                    predicted_scored_labels += shape.get_labels()

                add_saliency_maps_to_dataset_item(
                    dataset_item=dataset_item,
                    saliency_map=saliency_map,
                    model=self._task_environment.model,
                    labels=labels,
                    predicted_scored_labels=predicted_scored_labels,
                    explain_predicted_classes=explain_predicted_classes,
                    process_saliency_maps=process_saliency_maps,
                )

    def _det_add_predictions_to_dataset(self, all_results, width, height, confidence_threshold):
        shapes = []
        for label_idx, detections in enumerate(all_results):
            for i in range(detections.shape[0]):
                probability = float(detections[i, 4])
                coords = detections[i, :4].astype(float).copy()
                coords /= np.array([width, height, width, height], dtype=float)
                coords = np.clip(coords, 0, 1)

                if probability < confidence_threshold:
                    continue

                assigned_label = [ScoredLabel(self._labels[label_idx], probability=probability)]
                if coords[3] - coords[1] <= 0 or coords[2] - coords[0] <= 0:
                    continue

                shapes.append(
                    Annotation(
                        Rectangle(x1=coords[0], y1=coords[1], x2=coords[2], y2=coords[3]),
                        labels=assigned_label,
                    )
                )
        return shapes

    def _ins_seg_add_predictions_to_dataset(self, all_results, width, height, confidence_threshold):
        shapes = []
        for label_idx, (boxes, masks) in enumerate(zip(*all_results)):
            for mask, probability in zip(masks, boxes[:, 4]):
                if isinstance(mask, dict):
                    mask = mask_util.decode(mask)
                mask = mask.astype(np.uint8)
                probability = float(probability)
                contours, hierarchies = cv2.findContours(mask, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE)
                if hierarchies is None:
                    continue
                for contour, hierarchy in zip(contours, hierarchies[0]):
                    if hierarchy[3] != -1:
                        continue
                    if len(contour) <= 2 or probability < confidence_threshold:
                        continue
                    if self._task_type == TaskType.INSTANCE_SEGMENTATION:
                        points = [Point(x=point[0][0] / width, y=point[0][1] / height) for point in contour]
                    else:
                        box_points = cv2.boxPoints(cv2.minAreaRect(contour))
                        points = [Point(x=point[0] / width, y=point[1] / height) for point in box_points]
                    labels = [ScoredLabel(self._labels[label_idx], probability=probability)]
                    polygon = Polygon(points=points)
                    if cv2.contourArea(contour) > 0 and polygon.get_area() > 1e-12:
                        shapes.append(Annotation(polygon, labels=labels, id=ID(f"{label_idx:08}")))
        return shapes

    def _add_explanations_to_dataset(
        self, detections, explain_results, dataset, process_saliency_maps, explain_predicted_classes
    ):
        """Add saliency map to the dataset."""
        for dataset_item, detection, saliency_map in zip(dataset, detections, explain_results):
            labels = self._labels.copy()
            if saliency_map.shape[0] == len(labels) + 1:
                # Include the background as the last category
                labels.append(LabelEntity("background", Domain.DETECTION))

            shapes = self._get_shapes(detection, dataset_item.width, dataset_item.height, 0.4)
            predicted_scored_labels = []
            for shape in shapes:
                predicted_scored_labels += shape.get_labels()

            add_saliency_maps_to_dataset_item(
                dataset_item=dataset_item,
                saliency_map=saliency_map,
                model=self._task_environment.model,
                labels=labels,
                predicted_scored_labels=predicted_scored_labels,
                explain_predicted_classes=explain_predicted_classes,
                process_saliency_maps=process_saliency_maps,
            )

    @staticmethod
    def _update_anchors(origin, new):
        logger.info("Updating anchors")
        origin["heights"] = new["heights"]
        origin["widths"] = new["widths"]

    def _initialize_post_hook(self, options=None):
        super()._initialize_post_hook(options)
        options["model_builder"] = build_detector

        # if self._anchors are set somewhere, anchors had already been clusted
        # by this method or by loading trained model
        if should_cluster_anchors(self._model_cfg) and len(self._anchors) == 0:
            otx_dataset = get_configs_by_keys(self._data_cfg.data.train, "otx_dataset")
            assert len(otx_dataset) == 1
            otx_dataset = otx_dataset[0]
            cluster_anchors(
                self._model_cfg,
                self._recipe_cfg,
                otx_dataset,
            )
            self._update_anchors(self._anchors, self._model_cfg.model.bbox_head.anchor_generator)



"""Utils for common OTX algorithms."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
import glob
import os
import tempfile
from collections.abc import Mapping
from typing import Any, Dict, List, Tuple, Union

from mmcv import Config, ConfigDict

from otx.api.entities.datasets import DatasetEntity
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)
from otx.mpa.utils.logger import get_logger

from ._config_utils_get_configs_by_keys import get_configs_by_keys
from ._config_utils_get_configs_by_pairs import get_configs_by_pairs

logger = get_logger()


@check_input_parameters_type()
def remove_from_config(config: Union[Config, ConfigDict], key: str):
    """Update & Remove configs."""
    if key in config:
        if isinstance(config, Config):
            del config._cfg_dict[key]  # pylint: disable=protected-access
        elif isinstance(config, ConfigDict):
            del config[key]
        else:
            raise ValueError(f"Unknown config type {type(config)}")


@check_input_parameters_type()
def remove_from_configs_by_type(configs: List[ConfigDict], type_name: str):
    """Update & remove by type."""
    indices = []
    for i, config in enumerate(configs):
        type_name_ = config.get("type", None)
        if type_name_ == type_name:
            indices.append(i)
    for i in reversed(indices):
        configs.pop(i)


def update_config(
    config: Union[Config, ConfigDict],
    pairs: Dict[Tuple[Any, ...], Any],
):
    """Update configs by path as a key and value as a target."""
    for path, value in pairs.items():
        path_ = list(reversed(path))
        ptr = config
        key = None
        while path_:
            key = path_.pop()
            if isinstance(ptr, (Config, Mapping)):
                if key not in ptr:
                    ptr[key] = ConfigDict()
            elif isinstance(ptr, (list, tuple)):
                assert isinstance(key, int), f"{key} of {path} must be int for ({type(ptr)}: {ptr})"
                assert len(ptr) < key, f"{key} of {path} exceeds {len(ptr)}"
            if len(path_) == 0:
                ptr[key] = value
            ptr = ptr[key]


@check_input_parameters_type()
def get_dataset_configs(config: Union[Config, ConfigDict], subset: str) -> List[ConfigDict]:
    """Get 'datasets' configs."""
    if config.data.get(subset, None) is None:
        return []
    data_cfg = config.data[subset]
    data_cfgs = get_configs_by_keys(data_cfg, ["dataset", "datasets"])
    return data_cfgs if data_cfgs else [data_cfg]


@check_input_parameters_type({"dataset": DatasetParamTypeCheck})
def prepare_for_testing(config: Union[Config, ConfigDict], dataset: DatasetEntity) -> Config:
    """Prepare configs for testing phase."""
    config = copy.deepcopy(config)
    # FIXME. Should working directories be modified here?
    config.data.test.otx_dataset = dataset
    return config


@check_input_parameters_type()
def is_epoch_based_runner(runner_config: ConfigDict):
    """Check Epoch based or Iter based runner."""
    return "Epoch" in runner_config.type


@check_input_parameters_type()
def config_from_string(config_string: str) -> Config:
    """Generate an mmcv config dict object from a string.

    :param config_string: string to parse
    :return config: configuration object
    """
    with tempfile.NamedTemporaryFile("w", suffix=".py") as temp_file:
        temp_file.write(config_string)
        temp_file.flush()
        return Config.fromfile(temp_file.name)


@check_input_parameters_type()
def patch_default_config(config: Config):
    """Patch default config."""
    if "runner" not in config:
        config.runner = ConfigDict({"type": "EpochBasedRunner"})
    if "log_config" not in config:
        config.log_config = ConfigDict()
    if "evaluation" not in config:
        config.evaluation = ConfigDict()
    if "checkpoint_config" not in config:
        config.checkpoint_config = ConfigDict({"type": "CheckpointHook", "interval": 1})


@check_input_parameters_type()
def patch_data_pipeline(config: Config, data_pipeline: str = ""):
    """Replace data pipeline to data_pipeline.py if it exist."""
    if os.path.isfile(data_pipeline):
        data_pipeline_cfg = Config.fromfile(data_pipeline)
        config.merge_from_dict(data_pipeline_cfg)
    else:
        raise FileNotFoundError(f"data_pipeline: {data_pipeline} not founded")


@check_input_parameters_type()
def patch_color_conversion(config: Config):
    """Patch color conversion."""
    assert "data" in config

    for cfg in get_configs_by_pairs(config.data, dict(type="Normalize")):
        to_rgb = False
        if "to_rgb" in cfg:
            to_rgb = cfg.to_rgb
        cfg.to_rgb = not bool(to_rgb)


@check_input_parameters_type()
def patch_runner(config: Config):
    """Patch runner."""
    assert "runner" in config

    # Check that there is no conflict in specification of number of training epochs.
    # Move global definition of epochs inside runner config.
    if "total_epochs" in config:
        if is_epoch_based_runner(config.runner):
            if config.runner.max_epochs != config.total_epochs:
                logger.warning("Conflicting declaration of training epochs number.")
            config.runner.max_epochs = config.total_epochs
        else:
            logger.warning(f"Total number of epochs set for an iteration based runner {config.runner.type}.")
        remove_from_config(config, "total_epochs")

    # Change runner's type.
    if is_epoch_based_runner(config.runner):
        logger.info(f"Replacing runner from {config.runner.type} to EpochRunnerWithCancel.")
        config.runner.type = "EpochRunnerWithCancel"
    else:
        logger.info(f"Replacing runner from {config.runner.type} to IterBasedRunnerWithCancel.")
        config.runner.type = "IterBasedRunnerWithCancel"


@check_input_parameters_type()
def align_data_config_with_recipe(data_config: ConfigDict, config: Union[Config, ConfigDict]):
    """Align data_cfg with recipe_cfg."""
    # we assumed config has 'otx_dataset' and 'labels' key in it
    # by 'patch_datasets' function

    data_config = data_config.data
    config = config.data
    for subset in data_config.keys():
        subset_config = data_config.get(subset, {})
        for key in list(subset_config.keys()):
            found_config = get_configs_by_keys(config.get(subset), key, return_path=True)
            assert len(found_config) == 1
            value = subset_config.pop(key)
            path = list(found_config.keys())[0]
            update_config(subset_config, {path: value})


DEFAULT_META_KEYS = (
    "filename",
    "ori_filename",
    "ori_shape",
    "img_shape",
    "pad_shape",
    "scale_factor",
    "flip",
    "flip_direction",
    "img_norm_cfg",
)


def get_meta_keys(pipeline_step):
    """Update meta_keys for ignore_labels."""
    meta_keys = list(pipeline_step.get("meta_keys", DEFAULT_META_KEYS))
    meta_keys.append("ignored_labels")
    pipeline_step["meta_keys"] = set(meta_keys)
    return pipeline_step


@check_input_parameters_type()
def prepare_work_dir(config: Union[Config, ConfigDict]) -> str:
    """Prepare configs of working directory."""
    base_work_dir = config.work_dir
    checkpoint_dirs = glob.glob(os.path.join(base_work_dir, "checkpoints_round_*"))
    train_round_checkpoint_dir = os.path.join(base_work_dir, f"checkpoints_round_{len(checkpoint_dirs)}")
    os.makedirs(train_round_checkpoint_dir)
    config.work_dir = train_round_checkpoint_dir
    if "meta" not in config.runner:
        config.runner.meta = ConfigDict()
    config.runner.meta.exp_name = f"train_round_{len(checkpoint_dirs)}"
    return train_round_checkpoint_dir


@check_input_parameters_type()
def get_data_cfg(config: Union[Config, ConfigDict], subset: str = "train") -> Config:
    """Return dataset configs."""
    data_cfg = config.data[subset]
    while "dataset" in data_cfg:
        data_cfg = data_cfg.dataset
    return data_cfg



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import List

import torch

from otx.mpa.utils.logger import get_logger

from ..ops import OPS, Operation
from ..ops.infrastructures import ConstantV0
from .graph import Graph

logger = get_logger()


def get_constant_input_nodes(graph: Graph, node: Operation) -> List[Operation]:
    found = []
    for node in graph.predecessors(node):
        if node.type == "Constant":
            found.append(node)
    return found


def handle_merging_into_batchnorm(graph, type_patterns=[["Multiply", "Add"]], type_mappings=[{"gamma": 0, "beta": 1}]):

    assert len(type_patterns) == len(type_mappings)
    batchnorm_cls = OPS.get_by_type_version("BatchNormInference", 0)
    constant_cls = OPS.get_by_type_version("Constant", 0)

    for node in list(graph.nodes.keys()):
        if node not in graph:
            continue
        nodes = []
        type_mapping = {}
        for pattern_idx, type_pattern in enumerate(type_patterns):
            type_mapping = type_mappings[pattern_idx]
            nodes = graph.get_nodes_by_type_pattern(type_pattern, node)
            if nodes and len(nodes) == 1:
                nodes = nodes[0]
                break
        if not nodes:
            continue

        is_normalize = False
        for normalize_nodes in graph._normalize_nodes:
            if set(nodes).intersection(normalize_nodes):
                is_normalize = True
                break
        if is_normalize:
            logger.info(
                f"Skip merging {[i.name for i in nodes]} "
                f"becuase they are part of normalization (preprocessing of IR)"
            )
            continue

        shapes = []
        constants = []
        is_valid = True
        for node in nodes:
            constant = get_constant_input_nodes(graph, node)
            if len(constant) != 1:
                is_valid = False
                break
            constant = constant[0]
            shapes.append(constant.shape)
            constants.append(constant)
        if not is_valid:
            logger.info(
                f"Skip merging {[i.name for i in nodes]} " f"becuase it has more than one weights for node {node.name}."
            )
            continue

        if len(set(shapes)) != 1:
            logger.info(
                f"Skip merging {[i.name for i in nodes]} " f"becuase shape of weights are not the same. ({shapes})"
            )
            continue

        if len(set(shapes[0][2:])) != 1 or shapes[0][2] != 1:
            logger.info(f"Skip merging {[i.name for i in nodes]} " f"becuase shape of weights are not 1. ({shapes})")
            continue

        channel_dim = shapes[0][1]

        name = nodes[0].name + "/merged_bn"
        batchnorm = batchnorm_cls(name, shape=node.shape, epsilon=1e-10)

        gamma = (
            constants[type_mapping["gamma"]].data.squeeze() if "gamma" in type_mapping else torch.ones([channel_dim])
        )
        gamma = constant_cls(
            gamma,
            batchnorm.name + "/gamma",
            shape=((channel_dim,),),
            is_parameter=True,
        )

        beta = constants[type_mapping["beta"]].data.squeeze() if "beta" in type_mapping else torch.zeros([channel_dim])
        beta = constant_cls(
            beta,
            batchnorm.name + "/beta",
            shape=((channel_dim,),),
            is_parameter=True,
        )

        running_mean = (
            constants[type_mapping["running_mean"]].data.squeeze()
            if "running_mean" in type_mapping
            else torch.zeros([channel_dim])
        )
        running_mean = constant_cls(
            running_mean,
            batchnorm.name + "/running_mean",
            shape=((channel_dim,),),
            is_parameter=False,
        )

        running_variance = (
            constants[type_mapping["running_variance"]].data.squeeze()
            if "running_variance" in type_mapping
            else torch.ones([channel_dim])
        )
        running_variance = constant_cls(
            running_variance,
            batchnorm.name + "/running_variance",
            shape=((channel_dim,),),
            is_parameter=False,
        )

        logger.info(f"Merge {[i.name for i in nodes]} into batch normalization.")
        edges = []
        for predecessor in graph.predecessors(nodes[0]):
            if predecessor.type != "Constant":
                edges_attrs = graph.get_edge_data(predecessor, nodes[0])
                assert len(edges_attrs) == 1
                for edge_attrs in edges_attrs:
                    edges.append({"node_from": predecessor, "node_to": batchnorm, **edge_attrs})
        for successor in graph.successors(nodes[-1]):
            edges_attrs = graph.get_edge_data(nodes[-1], successor)
            assert len(edges_attrs) == 1
            for edge_attrs in edges_attrs:
                edges.append({"node_from": batchnorm, "node_to": successor, **edge_attrs})
        for node in nodes:
            graph.remove_node(node)
        for edge in edges:
            graph.add_edge(**edge)
        graph.add_edge(gamma, batchnorm)
        graph.add_edge(beta, batchnorm)
        graph.add_edge(running_mean, batchnorm)
        graph.add_edge(running_variance, batchnorm)


def handle_paired_batchnorm(graph, replace: bool = False, types: List[str] = ["Convolution", "GroupConvolution"]):
    batchnorm_cls = OPS.get_by_type_version("BatchNormInference", 0)
    constant_cls = OPS.get_by_type_version("Constant", 0)

    for node in list(graph.nodes.keys()):
        if node.type not in types:
            continue

        # if input is 1x1x... this node is probably in Squeeze-and-exitation network
        input_node, edge = list(graph.predecessors(node, True))[0]
        assert len(edge) == 1
        edge = edge[0]
        input_shape = input_node.shape[edge["out_port"]][2:]
        if len(set(input_shape)) == 1 and input_shape[0] == 1:
            logger.info(
                f"Skip a paired batch normalization for {node.name} " f"becuase input shape to it is {input_shape}."
            )
            continue

        bias_node = [n for n in graph.successors(node) if n.type == "Add"]
        if len(bias_node) == 1:
            bias_node = bias_node[0]
        else:
            bias_node = None

        # if bias node is not found we do not need to add batchnorm
        if bias_node is None:
            logger.info(f"Skip a paired batch normalization for {node.name} " "becuase it has no bias add node.")
            continue
        # if add node is not bias add node
        elif not isinstance(list(graph.predecessors(bias_node))[1], ConstantV0):
            logger.info(
                f"Skip a pared batch normalization for {node.name } "
                f"because {bias_node.name} is not a bias add node."
            )
            continue

        node_name = node.name
        channel_dim = node.attrs.shape[0][1]

        batchnorm = batchnorm_cls(node_name + "/paried_bn", shape=node.shape, epsilon=1e-10)

        gamma = torch.ones([channel_dim])
        gamma = constant_cls(
            batchnorm.name + "/gamma",
            data=gamma,
            shape=((channel_dim,),),
            is_parameter=True,
        )
        if replace and bias_node is not None:
            beta = list(graph.predecessors(bias_node))[1].data.squeeze()
        else:
            beta = torch.zeros([channel_dim])
        beta = constant_cls(
            batchnorm.name + "/beta",
            data=beta,
            shape=((channel_dim,),),
            is_parameter=True,
        )
        running_mean = torch.zeros([channel_dim])
        running_mean = constant_cls(
            batchnorm.name + "/running_mean",
            data=running_mean,
            shape=((channel_dim,),),
            is_parameter=False,
        )
        running_variance = torch.ones([channel_dim])
        running_variance = constant_cls(
            batchnorm.name + "/running_variance",
            data=running_variance,
            shape=((channel_dim,),),
            is_parameter=False,
        )

        if replace and bias_node is not None:
            logger.info(f"Replace {bias_node.name} with a paired batch normalization.")
            edges = []
            for successor in graph.successors(bias_node):
                edges_attrs = graph.get_edge_data(bias_node, successor)
                assert len(edges_attrs) == 1
                for edge_attrs in edges_attrs:
                    edges.append({"node_from": batchnorm, "node_to": successor, **edge_attrs})
            for predecessor in graph.predecessors(bias_node):
                if predecessor.type != "Constant":
                    edges_attrs = graph.get_edge_data(predecessor, bias_node)
                    assert len(edges_attrs) == 1
                    for edge_attrs in edges_attrs:
                        edges.append(
                            {
                                "node_from": predecessor,
                                "node_to": batchnorm,
                                **edge_attrs,
                            }
                        )
            graph.remove_node(bias_node)
            for edge in edges:
                graph.add_edge(**edge)
        else:
            logger.info(f"Append a paired batch normalization after {node.name}")
            edges = []
            for successor in graph.successors(node):
                edges_attrs = graph.get_edge_data(node, successor)
                assert len(edges_attrs) == 1
                for edge_attrs in edges_attrs:
                    edges.append({"node_from": batchnorm, "node_to": successor, **edge_attrs})
                graph.remove_edge(node, successor)
            for edge in edges:
                graph.add_edge(**edge)
            graph.add_edge(node, batchnorm)

        graph.add_edge(gamma, batchnorm)
        graph.add_edge(beta, batchnorm)
        graph.add_edge(running_mean, batchnorm)
        graph.add_edge(running_variance, batchnorm)


def handle_reshape(graph):

    for result in graph.get_nodes_by_types(["Result"]):
        for node in graph.predecessors(result):
            # some models, for example, dla-34, have reshape node as its predecessor
            # of result node and the reshape node reshapes the tensor to [1, -1]
            if node.type == "Reshape":
                input_node, shape = list(graph.predecessors(node))
                if torch.equal(shape.data, torch.tensor([1, -1])):
                    for shape_ in input_node.shape[0][::-1]:
                        if shape_ != 1:
                            break
                    logger.info(f"Change reshape to [-1, {shape_}]")
                    shape.data = torch.tensor([-1, shape_])



"""This module implements the Media entity."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc
from typing import Optional

import numpy as np

from otx.api.entities.annotation import Annotation


class IMediaEntity(metaclass=abc.ABCMeta):
    """Media entity interface.

    This interface is used to represent any kind of media data, on which users can annotate and tasks can perform
    training/analysis.
    """


class IMedia2DEntity(IMediaEntity, metaclass=abc.ABCMeta):
    """This interface is used to represent IMedia which is 2-dimensional media, i.e., containing height and width."""

    @property  # type:ignore
    @abc.abstractmethod
    def numpy(self) -> np.ndarray:
        """Returns the numpy representation of the 2D Media object."""
        raise NotImplementedError

    @numpy.setter  # type:ignore
    @abc.abstractmethod
    def numpy(self, value: np.ndarray):
        raise NotImplementedError

    @abc.abstractmethod
    def roi_numpy(self, roi: Optional[Annotation]) -> np.ndarray:
        """Returns the numpy representation of the 2D Media object while taking the roi into account."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def height(self) -> int:
        """Returns the height of the 2D Media object."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def width(self) -> int:
        """Returns the width representation of the 2D Media object."""
        raise NotImplementedError

    @property
    def path(self) -> Optional[str]:
        """Returns the path of the 2D Media object."""
        raise NotImplementedError



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from dataclasses import dataclass, field
from typing import Callable, List

from torch.nn import functional as F

from .builder import OPS
from .op import Attribute, Operation
from .utils import get_torch_padding


@dataclass
class MaxPoolV0Attribute(Attribute):
    strides: List[int]
    pads_begin: List[int]
    pads_end: List[int]
    kernel: List[int]
    rounding_type: str = field(default="floor")
    auto_pad: str = field(default="explicit")
    dilations: List[int] = field(default_factory=lambda: [])
    index_element_type: str = field(default="i64")
    axis: int = field(default=0)

    def __post_init__(self):
        super().__post_init__()
        valid_auto_pad = ["explicit", "same_upper", "same_Lower", "valid"]
        if self.auto_pad not in valid_auto_pad:
            raise ValueError(f"Invalid auto_pad {self.auto_pad}. " f"It must be one of {valid_auto_pad}.")
        valid_rounding_type = ["ceil", "floor"]
        if self.rounding_type not in valid_rounding_type:
            raise ValueError(
                f"Invalid rounding_type {self.rounding_type}. " f"It must be one of {valid_rounding_type}."
            )
        valid_index_element_type = ["i32", "i64"]
        if self.index_element_type not in valid_index_element_type:
            raise ValueError(
                f"Invalid index_element_type {self.index_element_type}. "
                f"It must be one of {valid_index_element_type}."
            )

        if not self.dilations:
            self.dilations = [1 for _ in self.strides]

        if self.axis != 0:
            raise NotImplementedError


@OPS.register()
class MaxPoolV0(Operation[MaxPoolV0Attribute]):
    TYPE = "MaxPool"
    VERSION = 0
    ATTRIBUTE_FACTORY = MaxPoolV0Attribute

    def forward(self, input):

        if input.dim() == 3:
            func = F.max_pool1d
        elif input.dim() == 4:
            func = F.max_pool2d
        elif input.dim() == 5:
            func = F.max_pool3d
        else:
            raise NotImplementedError

        padding = get_torch_padding(
            self.attrs.pads_begin,
            self.attrs.pads_end,
            self.attrs.auto_pad,
            list(input.shape[2:]),
            self.attrs.kernel,
            self.attrs.strides,
        )
        if isinstance(padding, Callable):
            input = padding(input=input)
            padding = 0

        return func(
            input=input,
            kernel_size=self.attrs.kernel,
            stride=self.attrs.strides,
            padding=padding,
            dilation=self.attrs.dilations,
            ceil_mode=self.attrs.rounding_type == "ceil",
            return_indices=True,
        )


@dataclass
class AvgPoolV1Attribute(Attribute):
    exclude_pad: bool
    strides: List[int]
    pads_begin: List[int]
    pads_end: List[int]
    kernel: List[int]
    rounding_type: str = field(default="floor")
    auto_pad: str = field(default="explicit")

    def __post_init__(self):
        super().__post_init__()
        valid_auto_pad = ["explicit", "same_upper", "same_Lower", "valid"]
        if self.auto_pad not in valid_auto_pad:
            raise ValueError(f"Invalid auto_pad {self.auto_pad}. " f"It must be one of {valid_auto_pad}.")
        valid_rounding_type = ["ceil", "floor"]
        if self.rounding_type not in valid_rounding_type:
            raise ValueError(
                f"Invalid rounding_type {self.rounding_type}. " f"It must be one of {valid_rounding_type}."
            )


@OPS.register()
class AvgPoolV1(Operation[AvgPoolV1Attribute]):
    TYPE = "AvgPool"
    VERSION = 1
    ATTRIBUTE_FACTORY = AvgPoolV1Attribute

    def __init__(self, *args, **kwargs):
        if "exclude-pad" in kwargs:
            kwargs["exclude_pad"] = kwargs.pop("exclude-pad")
        super().__init__(*args, **kwargs)

    def forward(self, input):
        if input.dim() == 3:
            func = F.avg_pool1d
        elif input.dim() == 4:
            func = F.avg_pool2d
        elif input.dim() == 5:
            func = F.avg_pool3d
        else:
            raise NotImplementedError

        padding = get_torch_padding(
            self.attrs.pads_begin,
            self.attrs.pads_end,
            self.attrs.auto_pad,
            list(input.shape[2:]),
            self.attrs.kernel,
            self.attrs.strides,
        )
        if isinstance(padding, Callable):
            input = padding(input=input)
            padding = 0

        return func(
            input=input,
            kernel_size=self.attrs.kernel,
            stride=self.attrs.strides,
            padding=padding,
            ceil_mode=self.attrs.rounding_type == "ceil",
            count_include_pad=not self.attrs.exclude_pad,
        )



"""Collection Pipeline for detection task."""
# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import copy
import tempfile
from typing import Any, Dict, Optional

import numpy as np
from mmdet.datasets.builder import PIPELINES

from otx.algorithms.common.utils.data import get_image
from otx.api.entities.label import Domain
from otx.api.utils.argument_checks import check_input_parameters_type

from .dataset import get_annotation_mmdet_format

_CACHE_DIR = tempfile.TemporaryDirectory(prefix="img-cache-")  # pylint: disable=consider-using-with


# pylint: disable=too-many-instance-attributes, too-many-arguments
@PIPELINES.register_module()
class LoadImageFromOTXDataset:
    """Pipeline element that loads an image from a OTX Dataset on the fly. Can do conversion to float 32 if needed.

    Expected entries in the 'results' dict that should be passed to this pipeline element are:
        results['dataset_item']: dataset_item from which to load the image
        results['dataset_id']: id of the dataset to which the item belongs
        results['index']: index of the item in the dataset

    :param to_float32: optional bool, True to convert images to fp32. defaults to False
    """

    @check_input_parameters_type()
    def __init__(self, to_float32: bool = False):
        self.to_float32 = to_float32

    @check_input_parameters_type()
    def __call__(self, results: Dict[str, Any]):
        """Callback function LoadImageFromOTXDataset."""
        # Get image (possibly from cache)
        img = get_image(results, _CACHE_DIR.name, to_float32=self.to_float32)
        shape = img.shape

        assert img.shape[0] == results["height"], f"{img.shape[0]} != {results['height']}"
        assert img.shape[1] == results["width"], f"{img.shape[1]} != {results['width']}"

        filename = f"Dataset item index {results['index']}"
        results["filename"] = filename
        results["ori_filename"] = filename
        results["img"] = img
        results["img_shape"] = shape
        results["ori_shape"] = shape
        # Set initial values for default meta_keys
        results["pad_shape"] = shape
        num_channels = 1 if len(shape) < 3 else shape[2]
        results["img_norm_cfg"] = dict(
            mean=np.zeros(num_channels, dtype=np.float32),
            std=np.ones(num_channels, dtype=np.float32),
            to_rgb=False,
        )
        results["img_fields"] = ["img"]

        return results


@PIPELINES.register_module()
class LoadAnnotationFromOTXDataset:
    """Pipeline element that loads an annotation from a OTX Dataset on the fly.

    Expected entries in the 'results' dict that should be passed to this pipeline element are:
        results['dataset_item']: dataset_item from which to load the annotation
        results['ann_info']['label_list']: list of all labels in the project
    """

    @check_input_parameters_type()
    def __init__(
        self,
        min_size: int,
        with_bbox: bool = True,
        with_label: bool = True,
        with_mask: bool = False,
        with_seg: bool = False,
        poly2mask: bool = True,
        with_text: bool = False,
        domain: Optional[Domain] = None,
    ):
        self.with_bbox = with_bbox
        self.with_label = with_label
        self.with_mask = with_mask
        self.with_seg = with_seg
        self.poly2mask = poly2mask
        self.with_text = with_text
        self.domain = domain
        self.min_size = min_size

    @staticmethod
    def _load_bboxes(results, ann_info):
        results["bbox_fields"].append("gt_bboxes")
        results["gt_bboxes"] = copy.deepcopy(ann_info["bboxes"])
        return results

    @staticmethod
    def _load_labels(results, ann_info):
        results["gt_labels"] = copy.deepcopy(ann_info["labels"])
        return results

    @staticmethod
    def _load_masks(results, ann_info):
        results["mask_fields"].append("gt_masks")
        results["gt_masks"] = copy.deepcopy(ann_info["masks"])
        return results

    @check_input_parameters_type()
    def __call__(self, results: Dict[str, Any]):
        """Callback function of LoadAnnotationFromOTXDataset."""
        dataset_item = results["dataset_item"]
        label_list = results["ann_info"]["label_list"]
        ann_info = get_annotation_mmdet_format(dataset_item, label_list, self.domain, self.min_size)
        if self.with_bbox:
            results = self._load_bboxes(results, ann_info)
            if results is None or len(results["gt_bboxes"]) == 0:
                return None
        if self.with_label:
            results = self._load_labels(results, ann_info)
        if self.with_mask:
            results = self._load_masks(results, ann_info)
        return results



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import json
import os
import tempfile

import pytest

from otx.algorithms.common.adapters.nncf.config import (
    compose_nncf_config,
    load_nncf_config,
    merge_dicts_and_lists_b_into_a,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@e2e_pytest_unit
def test_load_nncf_config():
    with pytest.raises(AssertionError):
        load_nncf_config("invalid_path")
    with tempfile.TemporaryDirectory() as directory:
        tmp = os.path.join(directory, "temp.json")
        with open(tmp, "w") as f:
            json.dump({"dummy": "dummy"}, f)
        assert {"dummy": "dummy"} == load_nncf_config(tmp)


@e2e_pytest_unit
def test_compose_nncf_config():
    nncf_config = {
        "base": {
            "find_unused_parameters": True,
            "nncf_config": {
                "target_metric_name": "mAP",
                "input_info": {"sample_size": [1, 3, 864, 864]},
                "compression": [],
                "log_dir": ".",
                "accuracy_aware_training": {
                    "mode": "early_exit",
                },
            },
        },
        "nncf_quantization": {
            "optimizer": {"lr": 0.0005},
            "nncf_config": {
                "compression": [
                    {
                        "algorithm": "quantization",
                        "initializer": {
                            "range": {"num_init_samples": 300},
                            "batchnorm_adaptation": {"num_bn_adaptation_samples": 300},
                        },
                    }
                ],
                "accuracy_aware_training": {
                    "mode": "early_exit",
                    "params": {
                        "maximal_absolute_accuracy_degradation": 0.01,
                        "maximal_total_epochs": 20,
                    },
                },
            },
        },
        "nncf_quantization_pruning": {
            "nncf_config": {
                "accuracy_aware_training": {
                    "mode": "adaptive_compression_level",
                    "params": {
                        "initial_training_phase_epochs": 5,
                        "maximal_total_epochs": 100,
                        "patience_epochs": 5,
                    },
                },
                "compression": [
                    {
                        "algorithm": "filter_pruning",
                        "ignored_scopes": ["{re}SingleStageDetector/SSDHead\\[bbox_head\\].*"],
                        "params": {
                            "schedule": "baseline",
                            "pruning_flops_target": 0.1,
                            "filter_importance": "geometric_median",
                        },
                    },
                    {
                        "algorithm": "quantization",
                        "initializer": {
                            "range": {"num_init_samples": 300},
                            "batchnorm_adaptation": {"num_bn_adaptation_samples": 300},
                        },
                    },
                ],
            }
        },
        "error": {"nncf_config": {"accuracy_aware_training": ["error"]}},
        "order_of_parts": ["nncf_quantization", "nncf_quantization_pruning", "error"],
    }

    assert nncf_config["base"] == compose_nncf_config(nncf_config, [])
    assert merge_dicts_and_lists_b_into_a(nncf_config["base"], nncf_config["nncf_quantization"]) == compose_nncf_config(
        nncf_config, ["nncf_quantization"]
    )
    assert merge_dicts_and_lists_b_into_a(
        nncf_config["base"], nncf_config["nncf_quantization_pruning"]
    ) == compose_nncf_config(nncf_config, ["nncf_quantization_pruning"])

    with pytest.raises(RuntimeError):
        compose_nncf_config(nncf_config, ["error"])


@e2e_pytest_unit
def test_merge_dicts_and_lists_b_into_a():
    assert {"a": 1, "b": 2} == merge_dicts_and_lists_b_into_a({"a": 1}, {"b": 2})
    assert [1, 2] == merge_dicts_and_lists_b_into_a([1], [2])
    assert {"a": [1, 2]} == merge_dicts_and_lists_b_into_a({"a": [1]}, {"a": [2]})



"""Configuration file of OTX Segmentation."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.


from sys import maxsize

from attr import attrs

from otx.algorithms.common.configs import (
    BaseConfig,
    LearningRateSchedule,
    POTQuantizationPreset,
)
from otx.api.configuration.elements import (
    add_parameter_group,
    boolean_attribute,
    configurable_boolean,
    configurable_float,
    configurable_integer,
    selectable,
    string_attribute,
)
from otx.api.configuration.model_lifecycle import ModelLifecycle

from .configuration_enums import Models

# pylint: disable=invalid-name


@attrs
class SegmentationConfig(BaseConfig):
    """Configurations of OTX Segmentation."""

    header = string_attribute("Configuration for an object detection task of MPA")
    description = header

    @attrs
    class __LearningParameters(BaseConfig.BaseLearningParameters):
        header = string_attribute("Learning Parameters")
        description = header

        learning_rate_fixed_iters = configurable_integer(
            default_value=100,
            min_value=0,
            max_value=5000,
            header="Number of iterations for fixed learning rate",
            description="",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        learning_rate_schedule = selectable(
            default_value=LearningRateSchedule.COSINE,
            header="Learning rate schedule",
            description="Specify learning rate scheduling for the MMDetection task. "
            "When training for a small number of epochs (N < 10), the fixed "
            "schedule is recommended. For training for 10 < N < 25 epochs, "
            "step-wise or exponential annealing might give better results. "
            "Finally, for training on large datasets for at least 20 "
            "epochs, cyclic annealing could result in the best model.",
            editable=True,
            visible_in_ui=True,
        )

    @attrs
    class __AlgoBackend(BaseConfig.BaseAlgoBackendParameters):
        header = string_attribute("Parameters for the MPA algo-backend")
        description = header

    @attrs
    class __Postprocessing(BaseConfig.BasePostprocessing):
        header = string_attribute("Postprocessing")
        description = header

        class_name = selectable(
            default_value=Models.BlurSegmentation,
            header="Model class for inference",
            description="Model classes with defined pre- and postprocessing",
            editable=False,
            visible_in_ui=True,
        )
        blur_strength = configurable_integer(
            header="Blur strength",
            description="With a higher value, the segmentation output will be smoother, but less accurate.",
            default_value=1,
            min_value=1,
            max_value=25,
            affects_outcome_of=ModelLifecycle.INFERENCE,
        )
        soft_threshold = configurable_float(
            default_value=0.5,
            header="Soft threshold",
            description="The threshold to apply to the probability output of the model, for each pixel. A higher value "
            "means a stricter segmentation prediction.",
            min_value=0.0,
            max_value=1.0,
            affects_outcome_of=ModelLifecycle.INFERENCE,
        )

    @attrs
    class __POTParameter(BaseConfig.BasePOTParameter):
        header = string_attribute("POT Parameters")
        description = header
        visible_in_ui = boolean_attribute(False)

        stat_subset_size = configurable_integer(
            header="Number of data samples",
            description="Number of data samples used for post-training optimization",
            default_value=300,
            min_value=1,
            max_value=maxsize,
        )

        preset = selectable(
            default_value=POTQuantizationPreset.PERFORMANCE,
            header="Preset",
            description="Quantization preset that defines quantization scheme",
            editable=False,
            visible_in_ui=False,
        )

    @attrs
    class __NNCFOptimization(BaseConfig.BaseNNCFOptimization):
        header = string_attribute("Optimization by NNCF")
        description = header
        visible_in_ui = boolean_attribute(False)

        enable_quantization = configurable_boolean(
            default_value=True,
            header="Enable quantization algorithm",
            description="Enable quantization algorithm",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        enable_pruning = configurable_boolean(
            default_value=False,
            header="Enable filter pruning algorithm",
            description="Enable filter pruning algorithm",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        pruning_supported = configurable_boolean(
            default_value=False,
            header="Whether filter pruning is supported",
            description="Whether filter pruning is supported",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        maximal_accuracy_degradation = configurable_float(
            default_value=1.0,
            min_value=0.0,
            max_value=100.0,
            header="Maximum accuracy degradation",
            description="The maximal allowed accuracy metric drop",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

    learning_parameters = add_parameter_group(__LearningParameters)
    postprocessing = add_parameter_group(__Postprocessing)
    algo_backend = add_parameter_group(__AlgoBackend)
    nncf_optimization = add_parameter_group(__NNCFOptimization)
    pot_parameters = add_parameter_group(__POTParameter)
